/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.wildfly;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.HasTreatmentWillRequest;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.CreateTreatmentWillRequest;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.CreateTreatmentWillResponse;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.DeleteTreatmentWillRequest;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.DeleteTreatmentWillResponse;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.GetTreatmentWillRequest;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.GetTreatmentWillResponse;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.HasTreatmentWillResponse;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.UpdateTreatmentWillRequest;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16.UpdateTreatmentWillResponse;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelperV2;
import idws.IdwsIntegrationTestHelper;
import org.hl7.btrv2.TreatmentWillV2;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.Before;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.MethodSorters;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.ws.soap.soap11.Envelope;
import org.opensaml.ws.soap.soap11.Header;
import org.opensaml.xml.XMLObject;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.oxm.jaxb.Jaxb2Marshaller;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.ws.client.core.WebServiceMessageCallback;
import org.springframework.ws.client.core.WebServiceTemplate;
import org.springframework.ws.soap.SoapMessage;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.StringWriter;
import java.util.LinkedList;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

/**
 * IDWS integration tests
 */
@RunWith(SpringRunner.class)
@SpringBootTest
@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class TreatmentWillV2IDWSITCase extends AbstractITCase {

    private static final String TEST_FILE_PATH = "target/test-classes/";
    private final String TEST_CPRNR = "0501792275";
    private final String SAML_CONFIG_LOCATION = "../ltr-btr-service-wildfly/src/test/resources/wsproxy/localhost-saml-config";
    private final String KEY_STORE_FILE_LOCATION = "../ltr-btr-service-wildfly/src/test/resources/wsproxy/idws_testcerts/fmk_idws_test.p12";
    private final String SOSI_ALIAS_JKS_LOCATION = TEST_FILE_PATH + "TEST whitelisted SP SOSI alias.jks";
    private final String STS_TRUSTED_BOOTSTRAP_TOKEN_LOCATION = TEST_FILE_PATH + "TestTrustedIdpForBootstrapToken.p12";
    private final String SOSI_STS_URL = "http://test1.ekstern-test.nspop.dk:8080/sts/services/Bst2Idws";
    private final String RESPONSE_AUDIENCE_URI = "https://fmk";
    private final String KEY_STORE_TYPE = "pkcs12";
    private final String CERTIFICATE_ALIAS = "FMK_IDWS_TEST (funktionscertifikat)";
    private final String KEY_STORE_PASSWORD = "Test1234";

    private static final String jaxbContextPathBTR = "dk.sundhedsdatastyrelsen.behandlingstestamente._2020._03._16";

    @Autowired
    private TestHelperV2.HL7Helper hl7Helper;

    private String uri;
    private Jaxb2Marshaller jaxb2Marshaller;
    private WebServiceTemplate webServiceTemplate;

    private static final String personIdentifier = "0501792275";

    @Before
    public void before() {
        uri = super.getItAddress() + "/btr";

        // Create marshaller
        jaxb2Marshaller = new Jaxb2Marshaller();
        jaxb2Marshaller.setContextPaths(jaxbContextPathBTR);

        // Create WebServiceTemplate
        webServiceTemplate = new WebServiceTemplate();
        webServiceTemplate.setMarshaller(jaxb2Marshaller);
        webServiceTemplate.setUnmarshaller(jaxb2Marshaller);

        // Initialise IdwsIntegrationTestHelper
        IdwsIntegrationTestHelper.init(KEY_STORE_FILE_LOCATION, SAML_CONFIG_LOCATION, SOSI_ALIAS_JKS_LOCATION, STS_TRUSTED_BOOTSTRAP_TOKEN_LOCATION, SOSI_STS_URL, RESPONSE_AUDIENCE_URI, KEY_STORE_TYPE, KEY_STORE_PASSWORD, CERTIFICATE_ALIAS);
    }
    
    @Test
    public void step0_IDWSRequest_HasTreatmentWill() throws Exception {
        // Define action
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16#HasTreatmentWill";

        // Create request
        final HasTreatmentWillRequest request = hl7Helper.createHasTreatmentWillRequest(personIdentifier);

        // Marshal request to String for creating IDWS request
        final String payload = marshalRequestToString(request, jaxb2Marshaller);

        // Create SOAP opemsaml envelope
        Assertion assertion = IdwsIntegrationTestHelper.createAssertion(TEST_CPRNR);
        Envelope soapEnvelope = IdwsIntegrationTestHelper.getTrustSoapEnvelope(
                uri, assertion, payload, soapAction, false, IdwsIntegrationTestHelper.credential);

        // Get IDWS headers only as DOM Sources
        List<Source> headerList = getHeaderList(soapEnvelope.getHeader());

        // Create request callback that adds IDWS SOAP header to the SoapMessage
        final WebServiceMessageCallback requestCallback = getWebServiceMessageCallBackAddingSoapHeader(soapAction, headerList);

        // Make actual IDWS request using default marshaller and adding IDWS headers using callback
        final HasTreatmentWillResponse response =
                (HasTreatmentWillResponse) webServiceTemplate.marshalSendAndReceive(uri, request, requestCallback);
        
        // Assert
        // Test PersonIdentifier
        assertEquals(false, response.isWillExists());
        
    }

    @Test
    public void step1_IDWSRequest_CreateTreatmentWill() throws Exception {
        // Define action
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16#CreateTreatmentWill";

        // Create request
        final TreatmentWillV2 treatmentWill = hl7Helper.getTreatmentWillExample();
        final CreateTreatmentWillRequest request = hl7Helper.createCreateTreatmentWillRequest(personIdentifier, treatmentWill);

        // Marshal request to String for creating IDWS request
        final String payload = marshalRequestToString(request, jaxb2Marshaller);

        // Create SOAP opemsaml envelope
        Assertion assertion = IdwsIntegrationTestHelper.createAssertion(TEST_CPRNR);
        Envelope soapEnvelope = IdwsIntegrationTestHelper.getTrustSoapEnvelope(
                uri, assertion, payload, soapAction, false, IdwsIntegrationTestHelper.credential);

        // Get IDWS headers only as DOM Sources
        List<Source> headerList = getHeaderList(soapEnvelope.getHeader());

        // Create request callback that adds IDWS SOAP header to the SoapMessage
        final WebServiceMessageCallback requestCallback = getWebServiceMessageCallBackAddingSoapHeader(soapAction, headerList);

        // Make actual IDWS request using default marshaller and adding IDWS headers using callback
        final CreateTreatmentWillResponse response =
                (CreateTreatmentWillResponse) webServiceTemplate.marshalSendAndReceive(uri, request, requestCallback);
        
        // The response is empty, but should not be null
        assertNotNull(response);
    }

    @Test
    public void step2_IDWSRequest_GetTreatmentWill() throws Exception {
        // Define action
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16#GetTreatmentWill";

        // Create request
        final GetTreatmentWillRequest request = hl7Helper.createGetTreatmentWillRequest(personIdentifier);

        // Marshal request to String for creating IDWS request
        final String payload = marshalRequestToString(request, jaxb2Marshaller);

        // Create SOAP opemsaml envelope
        Assertion assertion = IdwsIntegrationTestHelper.createAssertion(TEST_CPRNR);
        Envelope soapEnvelope = IdwsIntegrationTestHelper.getTrustSoapEnvelope(
                uri, assertion, payload, soapAction, false, IdwsIntegrationTestHelper.credential);

        // Get IDWS headers only as DOM Sources
        List<Source> headerList = getHeaderList(soapEnvelope.getHeader());

        // Create request callback that adds IDWS SOAP header to the SoapMessage
        final WebServiceMessageCallback requestCallback = getWebServiceMessageCallBackAddingSoapHeader(soapAction, headerList);

        // Make actual IDWS request using default marshaller and adding IDWS headers using callback
        final GetTreatmentWillResponse response =
                (GetTreatmentWillResponse) webServiceTemplate.marshalSendAndReceive(uri, request, requestCallback);

        // Assert
        // Test PersonIdentifier and acceptances
        assertNotNull(response);
        assertEquals(personIdentifier, response.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension());
        List<POCDMT000040Entry> allEntries = getAllEntries(response);
        assertEquals(1, allEntries.size());
        assertEquals(1, allEntries.get(0).getObservation().getValues().size());
        
        TreatmentWillV2 treatmentWillElement = (TreatmentWillV2) allEntries.get(0).getObservation().getValues().get(0);
        assertNotNull(treatmentWillElement);
        
        assertEquals(TestHelperV2.NO_LIFE_PROLONGING_IF_DYING, treatmentWillElement.getNoLifeProlongingIfDying().isValue());

        assertEquals(TestHelperV2.NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().getAcceptanceNeeded());
        
        assertEquals(TestHelperV2.NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().getAcceptanceNeeded());
        
        assertEquals(TestHelperV2.NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoForcedTreatmentIfIncapable().getAcceptanceNeeded());
    }
    
    private List<POCDMT000040Entry> getAllEntries(GetTreatmentWillResponse response) {
        return response.getClinicalDocument().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries();
    }
    
    @Test
    public void step3_IDWSRequest_UpdateTreatmentWill() throws Exception {
        // Define action
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16#UpdateTreatmentWill";

        // Create request
        final TreatmentWillV2 treatmentWill = hl7Helper.getTreatmentWillExample();
        final UpdateTreatmentWillRequest request = hl7Helper.createUpdateTreatmentWillRequest(personIdentifier, treatmentWill);

        // Marshal request to String for creating IDWS request
        final String payload = marshalRequestToString(request, jaxb2Marshaller);

        // Create SOAP opemsaml envelope
        Assertion assertion = IdwsIntegrationTestHelper.createAssertion(TEST_CPRNR);
        Envelope soapEnvelope = IdwsIntegrationTestHelper.getTrustSoapEnvelope(
                uri, assertion, payload, soapAction, false, IdwsIntegrationTestHelper.credential);

        // Get IDWS headers only as DOM Sources
        List<Source> headerList = getHeaderList(soapEnvelope.getHeader());

        // Create request callback that adds IDWS SOAP header to the SoapMessage
        final WebServiceMessageCallback requestCallback = getWebServiceMessageCallBackAddingSoapHeader(soapAction, headerList);

        // Make actual IDWS request using default marshaller and adding IDWS headers using callback
        final UpdateTreatmentWillResponse response =
                (UpdateTreatmentWillResponse) webServiceTemplate.marshalSendAndReceive(uri, request, requestCallback);

        // The response is empty, but should not be null
        assertNotNull(response);
    }

    @Test
    public void step4_IDWSRequest_GetTreatmentWill() throws Exception {
        // Define action
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16#GetTreatmentWill";

        // Create request
        final GetTreatmentWillRequest request = hl7Helper.createGetTreatmentWillRequest(personIdentifier);

        // Marshal request to String for creating IDWS request
        final String payload = marshalRequestToString(request, jaxb2Marshaller);

        // Create SOAP opemsaml envelope
        Assertion assertion = IdwsIntegrationTestHelper.createAssertion(TEST_CPRNR);
        Envelope soapEnvelope = IdwsIntegrationTestHelper.getTrustSoapEnvelope(
                uri, assertion, payload, soapAction, false, IdwsIntegrationTestHelper.credential);

        // Get IDWS headers only as DOM Sources
        List<Source> headerList = getHeaderList(soapEnvelope.getHeader());

        // Create request callback that adds IDWS SOAP header to the SoapMessage
        final WebServiceMessageCallback requestCallback = getWebServiceMessageCallBackAddingSoapHeader(soapAction, headerList);

        // Make actual IDWS request using default marshaller and adding IDWS headers using callback
        final GetTreatmentWillResponse response =
                (GetTreatmentWillResponse) webServiceTemplate.marshalSendAndReceive(uri, request, requestCallback);
        
        // Assert
        // Test PersonIdentifier
        assertEquals(personIdentifier, response.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension());
        
    }
    
    @Test
    public void step5_IDWSRequest_HasTreatmentWill() throws Exception {
        // Define action
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16#HasTreatmentWill";

        // Create request
        final HasTreatmentWillRequest request = hl7Helper.createHasTreatmentWillRequest(personIdentifier);

        // Marshal request to String for creating IDWS request
        final String payload = marshalRequestToString(request, jaxb2Marshaller);

        // Create SOAP opemsaml envelope
        Assertion assertion = IdwsIntegrationTestHelper.createAssertion(TEST_CPRNR);
        Envelope soapEnvelope = IdwsIntegrationTestHelper.getTrustSoapEnvelope(
                uri, assertion, payload, soapAction, false, IdwsIntegrationTestHelper.credential);

        // Get IDWS headers only as DOM Sources
        List<Source> headerList = getHeaderList(soapEnvelope.getHeader());

        // Create request callback that adds IDWS SOAP header to the SoapMessage
        final WebServiceMessageCallback requestCallback = getWebServiceMessageCallBackAddingSoapHeader(soapAction, headerList);

        // Make actual IDWS request using default marshaller and adding IDWS headers using callback
        final HasTreatmentWillResponse response =
                (HasTreatmentWillResponse) webServiceTemplate.marshalSendAndReceive(uri, request, requestCallback);
        
        // Assert
        // Test PersonIdentifier
        assertEquals(true, response.isWillExists());
        
    }
    
    @Test
    public void step6_IDWSRequest_DeleteTreatmentWill() throws Exception {
        // Define action
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16#DeleteTreatmentWill";

        // Create request
        final DeleteTreatmentWillRequest request = hl7Helper.createDeleteTreatmentWillRequest(personIdentifier);

        // Marshal request to String for creating IDWS request
        final String payload = marshalRequestToString(request, jaxb2Marshaller);

        // Create SOAP opemsaml envelope
        Assertion assertion = IdwsIntegrationTestHelper.createAssertion(TEST_CPRNR);
        Envelope soapEnvelope = IdwsIntegrationTestHelper.getTrustSoapEnvelope(
                uri, assertion, payload, soapAction, false, IdwsIntegrationTestHelper.credential);

        // Get IDWS headers only as DOM Sources
        List<Source> headerList = getHeaderList(soapEnvelope.getHeader());

        // Create request callback that adds IDWS SOAP header to the SoapMessage
        final WebServiceMessageCallback requestCallback = getWebServiceMessageCallBackAddingSoapHeader(soapAction, headerList);

        // Make actual IDWS request using default marshaller and adding IDWS headers using callback
        final DeleteTreatmentWillResponse response =
                (DeleteTreatmentWillResponse) webServiceTemplate.marshalSendAndReceive(uri, request, requestCallback);

        // Assert

    }

    private List<Source> getHeaderList(Header header) {
        List<Source> domList = new LinkedList<>();
        List<XMLObject> children = header.getOrderedChildren();
        for (XMLObject child : children) {
            domList.add(new DOMSource(child.getDOM()));
        }
        return domList;
    }

    private static WebServiceMessageCallback getWebServiceMessageCallBackAddingSoapHeader(String soapAction, List<Source> sourceList) {
        return message -> {
            SoapMessage soapMessage = (SoapMessage) message;
            soapMessage.setSoapAction(soapAction);
            Result headerResult = soapMessage.getSoapHeader().getResult();
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            for (Source source : sourceList) {
                transformer.transform(source, headerResult);
            }
        };
    }

    private static String marshalRequestToString(Object request, Jaxb2Marshaller jaxb2Marshaller) {
        StringWriter sw = new StringWriter();
        Result result = new StreamResult(sw);
        jaxb2Marshaller.marshal(request, result);
        return sw.toString();
    }

    // Alternative way to make IDWS request without WebServiceTemplate
    /*@Test
    public void testCallOfIDWSService() throws Exception {
        init();

        String action = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#GetTreatmentWill";
        String payload = getMessageFromFile("GetTreatmentWillRequest_body.xml");
        assertNotNull(payload);

        Assertion assertion = IdwsIntegrationTestHelper.createAssertion(TEST_CPRNR);
        Envelope soapEnvelope = IdwsIntegrationTestHelper.getTrustSoapEnvelope(
                        uri, assertion, payload, action, false, IdwsIntegrationTestHelper.credential);

        Element responseEnvelopeElement = IdwsIntegrationTestHelper.call(uri, action, soapEnvelope);
        assertNotNull(responseEnvelopeElement);

        String response = nodeToString(responseEnvelopeElement);
        assertNotNull(response);
        System.out.println("Response: " + response);

        assertTrue(response.contains(TEST_CPRNR));
    }

    private String nodeToString(Node node) throws TransformerException {
        StringWriter sw = new StringWriter();
        Transformer t = TransformerFactory.newInstance().newTransformer();
        t.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "true");
        t.setOutputProperty(OutputKeys.INDENT, "true");
        t.transform(new DOMSource(node), new StreamResult(sw));
        return sw.toString();
    }

    private String getMessageFromFile(String s) throws IOException {
        File file = new File(REQUEST_FILE_PATH + "/" + s);
        byte[] encoded = Files.readAllBytes(Paths.get(file.getPath()));
        return new String(encoded, "UTF-8");
    }*/
}
