/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service;

import dk.sundhedsdatastyrelsen.livstestamente.service.util.Properties;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ResponseBody;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;

/**
 * DCC konfigurationsservice (DKS)
 */
@Controller
public class DKSController {

    @SuppressWarnings("unused") private static final Logger logger = LogManager.getLogger(DKSController.class);

    private static final String RESOURCE_NAME_LTR = "/dksconfig/ltr.xml";
    private static final String RESOURCE_NAME_BTR = "/dksconfig/btr.xml";
    private static final String RESOURCE_NAME_BTRv2 = "/dksconfig/btrV2.xml";
    private static final String DATETIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss'Z'";

    private final Properties properties;

    @Autowired
    public DKSController(Properties properties) {this.properties = properties;}

    @GetMapping(value = "/dksconfig/ltr", produces = "application/xml")
    @ResponseBody
    public ResponseEntity<String> dksConfigLTR() throws IOException {
        return getDksConfig(RESOURCE_NAME_LTR, properties.getDccLTREndpoint());
    }

    @GetMapping(value = "/dksconfig/btr", produces = "application/xml")
    @ResponseBody
    public ResponseEntity<String> dksConfigBTR() throws IOException {
        return getDksConfig(RESOURCE_NAME_BTR, properties.getDccBTREndpoint());
    }
    
    @GetMapping(value = "/dksconfig/btrV2", produces = "application/xml")
    @ResponseBody
    public ResponseEntity<String> dksConfigBTRv2() throws IOException {
        return getDksConfig(RESOURCE_NAME_BTRv2, properties.getDccBTREndpoint());
    }

    private  ResponseEntity<String> getDksConfig(String resourceName, String dccEndpoint) throws IOException {
        URL url = getClass().getResource(resourceName);
        
        ZonedDateTime zdt = ZonedDateTime.ofInstant(Instant.ofEpochMilli(url.openConnection().getLastModified()), ZoneId.systemDefault());
        DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern(DATETIME_FORMAT);
        String fileTimestamp = dateTimeFormatter.format(zdt);

        InputStream in = getClass().getResourceAsStream(resourceName);
        BufferedReader reader = new BufferedReader(new InputStreamReader(in));
        StringBuilder out = new StringBuilder();
        String line;
        while ((line = reader.readLine()) != null) {
            out.append(line);
        }

        HttpHeaders headers = new HttpHeaders();
        headers.add(HttpHeaders.CACHE_CONTROL, "no-cache");

        return ResponseEntity.ok().headers(headers).body(getStringWithResolvedVariables(out, fileTimestamp, dccEndpoint));
    }

    private String getStringWithResolvedVariables(StringBuilder template, String filetimestamp, String dccEndpoint) {
        int timestampIndex = template.indexOf("TIMESTAMP");
        template = template.replace(timestampIndex, timestampIndex + 9, filetimestamp);

        return template.toString().replace("ENDPOINT_PLACEHOLDER", dccEndpoint);
    }
}
