/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.auditlog;

import dk.sds.dgws_idws_proxy.security.ProxyData;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.proxydata.ProxyMetaData;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;

public class AuditLogger {

    private static final String AUDIT_LOG_PREFIX = "audit.";
    private static final String DATE_FORMAT = "yyyy-MM-dd'T'HH:mm:ss.SSSZ";

    private Logger auditLog;

    public AuditLogger(Class<?> clazz) {
        auditLog = LogManager.getLogger(AUDIT_LOG_PREFIX + clazz.getName());
    }

    public void auditOk(RequestContextVO context, ProxyData proxyData) {
        long endTme = System.currentTimeMillis();
        if (proxyData != null) {
            info(context, proxyData, endTme - context.getTime().toEpochMilli());
        }
    }

    public void auditError(ProxyMetaData.RequestMetaData requestMetaData, Throwable error) {
        long endTme = System.currentTimeMillis();
        if (requestMetaData != null && requestMetaData.getProxyData() != null) {
            error(requestMetaData.getContext(), requestMetaData.getProxyData(), endTme - requestMetaData.getContext().getTime().toEpochMilli(), error.getMessage(), error);
        }
    }

    private void info(RequestContextVO context, ProxyData data, long durationMillis) {
        LogEntry entry = getAuditLogEntry(LogLevel.INFO);
        log(context, entry, data, durationMillis);
    }

    private void error(RequestContextVO context, ProxyData data, long durationMillis, String message, Throwable error) {
        LogEntry entry = getAuditLogEntry(LogLevel.ERROR, message, error);
        log(context, entry, data, durationMillis);
    }

    private void log(RequestContextVO context, LogEntry entry, ProxyData data, long durationMillis) {
        addParameter(entry, "requestId", data.getRequestId());
        addParameter(entry, "requestTimestamp", data.getRequestTimestamp());
        addParameter(entry, "requestType", data.getRequestType());
        addParameter(entry, "messageId", data.getMessageId());
        addParameter(entry, "flowId", data.getFlowId());
        addParameter(entry, "authenticationLevel", data.getAuthenticationLevel());
        addParameter(entry, "soapAction", data.getSoapAction());
        addParameter(entry, "cprnr", data.getCprnr());
        addParameter(entry, "givenName", data.getGivenName());
        addParameter(entry, "surName", data.getSurName());
        addParameter(entry, "authorizationCode", data.getAuthorizationCode());
        addParameter(entry, "occupation", data.getOccupation());
        addParameter(entry, "itSystemName", data.getItSystemName());
        addParameter(entry, "careProviderName", data.getCareProviderName());
        addParameter(entry, "careProviderID", data.getCareProviderID());
        addParameter(entry, "careProviderType", data.getCareProviderType());
        addParameter(entry, "durationOfCall", String.valueOf(durationMillis));
        addParameter(entry, "citizenCpr", context.getPersonIdentifier());

        entry.commit();
    }

    private void addParameter(LogEntry entry, String name, String value) {
        if (value != null) {
            entry.addParameter(name, value);
        }
    }

    private void addParameter(LogEntry entry, String name, Instant value) {
        if (value != null) {
        	DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern(DATE_FORMAT).withZone(ZoneId.systemDefault());
            addParameter(entry, name, dateTimeFormatter.format(value));
        }
    }
    
    private void addParameter(LogEntry entry, String name, Enum value) {
        if (value != null) {
            addParameter(entry, name, value.name());
        }
    }

    public LogEntry getAuditLogEntry(LogLevel logLevel) {
        return new LogEntry(auditLog, logLevel, null);
    }

    public LogEntry getAuditLogEntry(LogLevel logLevel, String message, Throwable error) {
        return new LogEntry(auditLog, logLevel, message, error);
    }

}
