/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import dk.cprsubscriber.dao.PersonSimplifiedDAO;
import dk.cprsubscriber.domain.PersonSimplifiedVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.PropertiesDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.PropertiesVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;

import java.text.ParseException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;

public abstract class AbstractDeleteJob {

    final PersonSimplifiedDAO personSimplifiedDAO;
    final PropertiesDAO propertiesDAO;

    public static final String LAST_DELETION_TIME = "LAST_DELETION_TIME";
    public static final String LAST_CPR_MAX_PID = "LAST_CPR_MAX_PID";
    public static final String STATUS_DECEASED = "90";

    protected AbstractDeleteJob(PersonSimplifiedDAO personSimplifiedDAO, PropertiesDAO propertiesDAO) {
        this.personSimplifiedDAO = personSimplifiedDAO;
        this.propertiesDAO = propertiesDAO;
    }

    boolean personIsDeceasedOneYearAgo(PersonSimplifiedVO person) {
        return STATUS_DECEASED.equals(person.getStatus());
    }

    Instant subtractOneYearFromNow(Instant instant) {
    	ZonedDateTime zonedDateTime = instant.atZone(ZoneId.systemDefault());
        zonedDateTime = zonedDateTime.minusYears(1);
        return zonedDateTime.toInstant();
    }

    void writeLatestRuntime(Instant currentRunStartedTime) throws DAOException {
        propertiesDAO.updateProperty(LAST_DELETION_TIME, currentRunStartedTime.toString());
    }

    void writeHighestPid(long maxPid) throws DAOException {
        propertiesDAO.updateProperty(LAST_CPR_MAX_PID, "" + maxPid);
    }

    Instant getLatestRunTime() throws DAOException, ParseException {
        PropertiesVO lastDeletionTime = propertiesDAO.getProperty(LAST_DELETION_TIME);
        if (lastDeletionTime == null || lastDeletionTime.getPropertyValue() == null) {
            throw new DAOException("Last deletion not found in database (key " + LAST_DELETION_TIME + ")");
        }
        return Instant.parse(lastDeletionTime.getPropertyValue());
    }

    long getLatestMaxPid() throws DAOException {
        PropertiesVO lastCprSyncTime = propertiesDAO.getProperty(LAST_CPR_MAX_PID);
        if (lastCprSyncTime == null || lastCprSyncTime.getPropertyValue() == null) {
            throw new DAOException("Max CPR PID not found in database (propertyKey " + LAST_CPR_MAX_PID + ")");
        }
        try {
            Long.valueOf(lastCprSyncTime.getPropertyValue());
        } catch (NumberFormatException e) {
            throw new DAOException("Unparsable long value from database (propertyKey " + LAST_CPR_MAX_PID + ", value=" + lastCprSyncTime.getPropertyValue() + ")");
        }
        return Long.valueOf(lastCprSyncTime.getPropertyValue());
    }
}
