/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import dk.cprsubscriber.dao.PersonSimplifiedDAO;
import dk.cprsubscriber.domain.PersonSimplifiedVO;
import dk.cprsubscriber.exceptions.CPRDAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.PropertiesDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.Properties;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.text.ParseException;
import java.time.Instant;
import java.util.List;

@Component
@EnableScheduling
public class DeleteLivingWillJob extends AbstractDeleteJob {
    private static final Logger logger = LogManager.getLogger(DeleteLivingWillJob.class);

    private final int delayWhenErrorMs = 10_000;
    private final int maximumWaitTimeMs = 5 * 60 * 1000;

    private final LivingWillDAO livingWillDAO;
    private final Properties properties;

    private int errorCount = 0;

    @Autowired
    public DeleteLivingWillJob(LivingWillDAO livingWillDAO,
                               PersonSimplifiedDAO personSimplifiedDAO,
                               @Qualifier("livingWillPropertiesDAOImpl") PropertiesDAO propertiesDAO,
                               Properties properties) {
        super(personSimplifiedDAO, propertiesDAO);
        this.livingWillDAO = livingWillDAO;
        this.properties = properties;
    }

    @Scheduled(cron = "${jobs.delete.schedule.ltr:}")
    public void runDeleteLivingWill() throws InterruptedException {
        if (!properties.isJobsDeleteLtrEnabled()) {
            logger.error("Delete Organ Donor Registration job is disabled. This must not happen in production.");
            return;
        }
        logger.debug("Deletion Job starting");

        try {
            Instant currentRunStartedTime = Instant.now();
            Instant oneYearAgoFromNow = subtractOneYearFromNow(currentRunStartedTime);

            Instant latestRuntimeSubtractedOneYear = getLatestRuntimeSubtractedOneYear();
            long latestMaxPid = getLatestMaxPid();

            List<PersonSimplifiedVO> updatedPersons = getPersonsUpdatedBetween(latestRuntimeSubtractedOneYear, oneYearAgoFromNow, latestMaxPid);

            int loops = 0;
            long maxPid = 0;
            while (updatedPersons.size() > 0 && loops++ < properties.getJobsDeleteLtrCprMaxLoops()) {
                long maxBatchPid = 0;
                for (PersonSimplifiedVO person : updatedPersons) {
                    if (person == null) {
                        continue;
                    }
                    if (person.getPid() > maxBatchPid) {
                        maxBatchPid = person.getPid();
                    }
                    // Person is deceased
                    boolean personExistsInIndex = personExistsInIndex(person);
                    logger.debug("Person " + person.getCpr() + " with status " + person.getStatus() + " updated. ExistsInIndex=" + personExistsInIndex);
                    if (personIsDeceasedOneYearAgo(person)) {
                        //person exists
                        if (personStillDecreased(person, currentRunStartedTime) && personExistsInIndex) {
                            cleanupPersonToBeDeleted(person);
                        }
                    }
                }
                maxPid = maxBatchPid;
                updatedPersons = getPersonsUpdatedBetween(latestRuntimeSubtractedOneYear, oneYearAgoFromNow, maxBatchPid);
            }

            writeLatestRuntime(currentRunStartedTime);
            writeHighestPid(maxPid);
            errorCount = 0;

        } catch (CPRDAOException | DAOException | ParseException e) {
            errorCount++;
            logger.error("Deletion Job unable to run because of an error.", e);
        } finally {
            // It probably does not help to try again immediately. So wait longer for every error.
            Thread.sleep(Math.min(errorCount * delayWhenErrorMs, maximumWaitTimeMs));
        }

        logger.debug("Deletion Job done");
    }

    public String getHealthInformation() throws ParseException, DAOException {
        String text = "";

        text += properties.isJobsDeleteLtrEnabled() ? "Enabled" : "Disabled";
        text += ", Last successful deletion: " + (getLatestRunTime() == null ? "Never" : getLatestRunTime().toString());
        text += ", Error count: " + errorCount;

        return text;
    }

    private Instant getLatestRuntimeSubtractedOneYear() throws DAOException, ParseException {
        Instant lastDeletionDate = getLatestRunTime();
        return subtractOneYearFromNow(lastDeletionDate);
    }

    private List<PersonSimplifiedVO> getPersonsUpdatedBetween(Instant latestRuntime, Instant startTime, long minPid) throws CPRDAOException {
        List<PersonSimplifiedVO> updated = personSimplifiedDAO.getPersonsUpdatedBetween(latestRuntime, startTime, minPid, properties.getJobsDeleteLtrCprMaxResults());
        logger.debug("getPersonsUpdatedBetween(" + latestRuntime + ", " + startTime + ", " + minPid + ") -> " + updated.size() + " records");
        return updated;
    }

    private boolean personStillDecreased(PersonSimplifiedVO person, Instant currentRunStartedTime) throws CPRDAOException {
        PersonSimplifiedVO currentPerson = personSimplifiedDAO.getRecordByDate(person.getCpr(), currentRunStartedTime);
        return personIsDeceasedOneYearAgo(currentPerson);
    }

    private boolean personExistsInIndex(PersonSimplifiedVO person) throws DAOException {
        return livingWillDAO.existsLivingWillRecords(person.getCpr());
    }

    private void cleanupPersonToBeDeleted(PersonSimplifiedVO person) throws DAOException {
        logger.debug("Now deleting organ donor registration for person with personId: [" + person.getCpr() + "]");
        livingWillDAO.deleteLivingWillRecords(person.getCpr());
    }
}
