/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import dk.sds.dgws_idws_proxy.security.ProxyData;
import dk.sundhedsdatastyrelsen.livstestamente.service.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.LivingWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.TreatmentWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultUtil;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.Properties;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.time.Year;
import java.time.ZoneOffset;

@Service
public class TreatmentWillServiceImpl implements TreatmentWillService {

    private static final Logger logger = LogManager.getLogger(TreatmentWillServiceImpl.class);

    private static AuditLogger auditLogger = new AuditLogger(TreatmentWillServiceImpl.class);

    static final String SERVICE_IDENTIFIER = "Behandlingstestamente";

    private final Properties properties;
    private final TreatmentWillDAO treatmentWillDAO;
    private final LivingWillDAO livingWillDAO;
    private final MinLogDataWrapper minLogDataWrapper;
    private final AccessControl accessControl;
    private final TreatmentWillValidator treatmentWillValidator;
    private final LivingWillValidator livingWillValidator;

    static final Instant VALID_TO_DEFAULT = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    static final int INITIAL_VERSION_DEFAULT = 1;

    @Autowired
    public TreatmentWillServiceImpl(
            Properties properties,
            TreatmentWillDAO treatmentWillDAO,
            LivingWillDAO livingWillDAO,
            MinLogDataWrapper minLogDataWrapper,
            @Qualifier("treatmentWillAccessControlImpl") AccessControl accessControl,
            TreatmentWillValidator treatmentWillValidator,
            LivingWillValidator livingWillValidator) {
        this.properties = properties;
        this.treatmentWillDAO = treatmentWillDAO;
        this.livingWillDAO = livingWillDAO;
        this.minLogDataWrapper = minLogDataWrapper;
        this.accessControl = accessControl;
        this.treatmentWillValidator = treatmentWillValidator;
        this.livingWillValidator = livingWillValidator;
    }

    @Override
    public TreatmentWillVO getTreatmentWill(ServiceActorVO actor, RequestContextVO context, ProxyData proxyData) {
        TreatmentWillVO result = null;

        accessControl.checkGetAccess(context);

        try {
            // Get any existing TreatmentWill
            final String personIdentifier = context.getPersonIdentifier();
            result = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, context.getTime());

            // Validate existing TreatmentWill
            treatmentWillValidator.validateTreatmentWillExists(result);

            auditLogger.auditOk(context, proxyData);

            // Log to MinLog. If MinLog fails, it throws an exception
            logReadActivityToMinLog(actor, context);
        } catch (DAOException | MinLogException e) {
            FaultUtil.throwFault(FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e);
        }

        return result;
    }

    @Override
    public boolean hasTreatmentWill(RequestContextVO context, ProxyData proxyData) {
        boolean result = false;

        accessControl.checkHasAccess(context);

        try {
            final String personIdentifier = context.getPersonIdentifier();
            final TreatmentWillVO currentTreatmentWill = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, context.getTime());
            result = currentTreatmentWill != null;

            auditLogger.auditOk(context, proxyData);

            // No minLog needed
        } catch (DAOException e) {
            FaultUtil.throwFault(FaultCode.HAS_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e);
        }
        return result;
    }

    @Override
    @Transactional("treatmentWillDataSourceTxManager")
    public void createTreatmentWill(TreatmentWillVO treatmentWillVO, ServiceActorVO actor, RequestContextVO context, ProxyData proxyData, boolean skipNoLifeProlongingIfDyingAcceptanceNeeded) {
        accessControl.checkCreateAccess(context);

        try {
            // Get any existing LivingWill and TreatmentWill
            final String personIdentifier = context.getPersonIdentifier();
            final LivingWillVO currentVersionOfLivingWill = livingWillDAO.getCurrentLivingWill(personIdentifier, context.getTime());
            final TreatmentWillVO currentVersionOfTreatmentWill = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, context.getTime());

            // Validate no existing LivingWill
            livingWillValidator.validateLivingWillNotExists(currentVersionOfLivingWill);

            // Validate no existing TreatmentWill
            treatmentWillValidator.validateTreatmentWillNotExists(currentVersionOfTreatmentWill);

            // No need to validate PersonIdentifier, as only the person themselves can create their own LivingWill

            treatmentWillVO.setVersion(INITIAL_VERSION_DEFAULT);
            treatmentWillVO.setValidFrom(context.getTime());
            treatmentWillVO.setValidTo(VALID_TO_DEFAULT);
            treatmentWillVO.setModifiedDateTime(context.getTime());

            auditLogger.auditOk(context, proxyData);
            if (skipNoLifeProlongingIfDyingAcceptanceNeeded) {
            	treatmentWillDAO.createTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded(treatmentWillVO);	
            } else {
            	treatmentWillDAO.createTreatmentWill(treatmentWillVO);
            }
            
            minLogIfAdminUser(context, actor);
        } catch (DAOException | MinLogException e) {
            FaultUtil.throwFault(FaultCode.CREATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e);
        }
    }

    @Override
    @Transactional("treatmentWillDataSourceTxManager")
    public void updateTreatmentWill(TreatmentWillVO treatmentWillVO, ServiceActorVO actor, RequestContextVO context, ProxyData proxyData, boolean skipNoLifeProlongingIfDyingAcceptanceNeeded) {
        accessControl.checkUpdateAccess(context);

        try {
            // Get any existing TreatmentWill
            final String personIdentifier = context.getPersonIdentifier();
            final TreatmentWillVO currentVersionOfTreatmentWill = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, context.getTime());

            // Validate existing TreatmentWill
            treatmentWillValidator.validateTreatmentWillExists(currentVersionOfTreatmentWill);

            final long currentPid = currentVersionOfTreatmentWill.getTreatmentWillPID();
            final int currentVersion = currentVersionOfTreatmentWill.getVersion();
            int nextVersion = currentVersion + 1;

            treatmentWillVO.setVersion(nextVersion);
            treatmentWillVO.setValidFrom(context.getTime());
            treatmentWillVO.setValidTo(VALID_TO_DEFAULT);
            treatmentWillVO.setModifiedDateTime(context.getTime());

            treatmentWillDAO.updateValidToTimestamp(currentPid, context.getTime());
            
            if (skipNoLifeProlongingIfDyingAcceptanceNeeded) {
            	treatmentWillDAO.createTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded(treatmentWillVO);	
            } else {
            	treatmentWillDAO.createTreatmentWill(treatmentWillVO);
            }
            auditLogger.auditOk(context, proxyData);
            minLogIfAdminUser(context, actor);
        } catch (DAOException | MinLogException e) {
            FaultUtil.throwFault(FaultCode.UPDATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e);
        }
    }

    @Override
    @Transactional("treatmentWillDataSourceTxManager")
    public void deleteTreatmentWill(ServiceActorVO actor, RequestContextVO context, ProxyData proxyData) {
        accessControl.checkDeleteAccess(context);

        try {
            // Get any existing TreatmentWill
            final String personIdentifier = context.getPersonIdentifier();
            final TreatmentWillVO currentVersionOfTreatmentWill = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, context.getTime());

            // Validate existing TreatmentWill
            treatmentWillValidator.validateTreatmentWillExists(currentVersionOfTreatmentWill);

            final long currentPid = currentVersionOfTreatmentWill.getTreatmentWillPID();
            treatmentWillDAO.updateValidToTimestamp(currentPid, context.getTime());
            auditLogger.auditOk(context, proxyData);
            minLogIfAdminUser(context, actor);
        } catch (DAOException | MinLogException e) {
            FaultUtil.throwFault(FaultCode.DELETE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e);
        }
    }

    private void minLogIfAdminUser(RequestContextVO context, ServiceActorVO actor) throws MinLogException {
        if (context.getUserType() == UserType.Admin) {
            logReadActivityToMinLog(actor, context);
        }
    }

    private void logReadActivityToMinLog(ServiceActorVO actor, RequestContextVO context) throws MinLogException {
        minLogDataWrapper.logReadActivity(
                context,
                properties.getMinLogReadActivityTextBTR(),
                context.getPersonIdentifier(),
                actor.getOrganizationId(),
                actor.getOrganizationName(),
                actor.getOrganizationIdSource() != null ? OrganizationTypeVO.valueOf(actor.getOrganizationIdSource()) : null,
                actor.getPersonIdentifier(),
                null,
                SERVICE_IDENTIFIER);
    }

}
