/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.conf;

import dk.sundhedsdatastyrelsen.livstestamente.service.util.CustomValidationInterceptor;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.Properties;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.web.servlet.ServletRegistrationBean;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.io.ClassPathResource;
import org.springframework.core.io.Resource;
import org.springframework.ws.config.annotation.EnableWs;
import org.springframework.ws.config.annotation.WsConfigurerAdapter;
import org.springframework.ws.server.EndpointInterceptor;
import org.springframework.ws.transport.http.MessageDispatcherServlet;
import org.springframework.ws.wsdl.WsdlDefinition;
import org.springframework.ws.wsdl.wsdl11.DefaultWsdl11Definition;
import org.springframework.xml.validation.XmlValidator;
import org.springframework.xml.validation.XmlValidatorFactory;
import org.springframework.xml.xsd.SimpleXsdSchema;
import org.springframework.xml.xsd.XsdSchema;
import org.springframework.xml.xsd.XsdSchemaCollection;

import java.io.IOException;
import java.util.List;

@EnableWs
@Configuration
public class WebServiceConfig extends WsConfigurerAdapter {

    private static final Logger logger = LogManager.getLogger(WebServiceConfig.class);

    private final dk.sundhedsdatastyrelsen.livstestamente.service.util.Properties properties;

    private final ClassPathResource xsdSchemaResource_2018_05_01_livstestamenteregister = new ClassPathResource("schema/2018_05_01/livstestamenteregister.xsd");
    private final ClassPathResource xsdSchemaResource_2018_05_01_behandlingstestamenteregister = new ClassPathResource("schema/2018_05_01/behandlingstestamenteregister.xsd");
    private final ClassPathResource xsdSchemaResource_2020_03_16_behandlingstestamenteregister = new ClassPathResource("schema/2020_03_16/behandlingstestamenteregister.xsd");
    
    @Autowired
    public WebServiceConfig(Properties properties) {
        this.properties = properties;
    }

    @Bean
    public ServletRegistrationBean<MessageDispatcherServlet> messageDispatcherServlet(ApplicationContext applicationContext) {
        MessageDispatcherServlet servlet = new MessageDispatcherServlet();
        servlet.setApplicationContext(applicationContext);
        servlet.setTransformWsdlLocations(true);
        return new ServletRegistrationBean<>(servlet, "/ltr", "/btr");
    }

    @Bean
    public WsdlDefinition wsdl11DefinitionLTR(XsdSchema xsdSchema_2018_05_01_livstestamenteregister) {
        DefaultWsdl11Definition wsdl11Definition = new DefaultWsdl11Definition();
        wsdl11Definition.setPortTypeName("LivingWillPort");
        wsdl11Definition.setLocationUri("/btr-service/ltr");
        wsdl11Definition.setTargetNamespace("http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01");
        wsdl11Definition.setSchema(xsdSchema_2018_05_01_livstestamenteregister);
        return wsdl11Definition;
    }

    @Bean
    public WsdlDefinition wsdl11DefinitionBTR(XsdSchema xsdSchema_2018_05_01_behandlingstestamenteregister) {
        DefaultWsdl11Definition wsdl11Definition = new DefaultWsdl11Definition();
        wsdl11Definition.setPortTypeName("TreatmentWillPort");
        wsdl11Definition.setLocationUri("/btr-service/btr");
        wsdl11Definition.setTargetNamespace("http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01");
        wsdl11Definition.setSchema(xsdSchema_2018_05_01_behandlingstestamenteregister);
        return wsdl11Definition;
    }
    
    @Bean
    public WsdlDefinition wsdl11DefinitionBTR2(XsdSchema xsdSchema_2020_03_16_behandlingstestamenteregister) {
        DefaultWsdl11Definition wsdl11Definition = new DefaultWsdl11Definition();
        wsdl11Definition.setPortTypeName("TreatmentWillPort");
        wsdl11Definition.setLocationUri("/btr-service/btr");
        wsdl11Definition.setTargetNamespace("http://sundhedsdatastyrelsen.dk/behandlingstestamente/2020/03/16");
        wsdl11Definition.setSchema(xsdSchema_2020_03_16_behandlingstestamenteregister);
        return wsdl11Definition;
    }

    @Override
    public void addInterceptors(List<EndpointInterceptor> interceptors) {
        final CustomValidationInterceptor validatingInterceptor = new CustomValidationInterceptor(properties);
        validatingInterceptor.setValidateRequest(properties.isSchemavalidationValidateRequests());
        validatingInterceptor.setValidateResponse(properties.isSchemavalidationValidateResponse());
        validatingInterceptor.setXsdSchemaCollection(ltrBtrSchemaCollection());
        interceptors.add(validatingInterceptor);
        super.addInterceptors(interceptors);
    }

    @Bean
    public XsdSchema xsdSchema_2018_05_01_livstestamenteregister() {
        return new SimpleXsdSchema(xsdSchemaResource_2018_05_01_livstestamenteregister);
    }

    @Bean
    public XsdSchema xsdSchema_2018_05_01_behandlingstestamenteregister() {
        return new SimpleXsdSchema(xsdSchemaResource_2018_05_01_behandlingstestamenteregister);
    }
    
    @Bean
    public XsdSchema xsdSchema_2020_03_16_behandlingstestamenteregister() {
        return new SimpleXsdSchema(xsdSchemaResource_2020_03_16_behandlingstestamenteregister);
    }

    @Bean
    public XsdSchemaCollection ltrBtrSchemaCollection() {
        return new XsdSchemaCollection() {
            @Override
            public XsdSchema[] getXsdSchemas() {
                return new XsdSchema[]{
                        xsdSchema_2018_05_01_livstestamenteregister(),
                        xsdSchema_2018_05_01_behandlingstestamenteregister(),
                        xsdSchema_2020_03_16_behandlingstestamenteregister()
                };
            }

            @Override
            public XmlValidator createValidator() {
                try {
                    Resource[] schemas = new Resource[]{
                            xsdSchemaResource_2018_05_01_livstestamenteregister,
                            xsdSchemaResource_2018_05_01_behandlingstestamenteregister, 
                            xsdSchemaResource_2020_03_16_behandlingstestamenteregister
                    };
                    return XmlValidatorFactory.createValidator(schemas, "http://www.w3.org/2001/XMLSchema");
                } catch (IOException e) {
                    logger.error("Error creating XML Schema validators", e);
                }
                return null;
            }
        };
    }
}
