/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.dao.impl;

import dk.sundhedsdatastyrelsen.livstestamente.service.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.impl.abstr.AbstractDAOImpl;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.support.DataAccessUtils;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.stereotype.Repository;

import javax.sql.DataSource;
import java.sql.Types;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Objects;

@SuppressWarnings("SqlResolve")
@Repository
public class LivingWillDAOImpl extends AbstractDAOImpl implements LivingWillDAO {

    @Autowired
    public LivingWillDAOImpl(DataSource livingWillDataSource) {
        super(livingWillDataSource);
    }

    @Override
    public LivingWillVO getCurrentLivingWill(String personIdentifier, Instant validAt) throws DAOException {
        try {
            final String sql = "SELECT l.* FROM LivingWill l " +
                    "WHERE l.PersonIdentifier = :PersonIdentifier " +
                    "AND l.ValidFrom <= :ValidAt AND :ValidAt < l.ValidTo " +
                    "ORDER BY l.ValidTo DESC LIMIT 1";
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("PersonIdentifier", personIdentifier, Types.VARCHAR)
                    .addValue("ValidAt", java.sql.Timestamp.from(validAt.truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP);
            return DataAccessUtils.singleResult(namedParamJdbcTemplate.query(sql, parameterSource, new LivingWillRowMapper()));
        } catch (DataAccessException e) {
            throw new DAOException("An error occurred reading LivingWill", e);
        }
    }

    @Override
    public boolean existsLivingWillRecords(String personIdentifier) throws DAOException {
        try {
            final String sql = "SELECT COUNT(*) FROM LivingWill l " +
                    "WHERE l.PersonIdentifier = :PersonIdentifier";
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("PersonIdentifier", personIdentifier, Types.VARCHAR);
            final long count = Objects.requireNonNull(namedParamJdbcTemplate.queryForObject(sql, parameterSource, Long.class));
            return count > 0;
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred counting LivingWill", e);
        }
    }

    @Override
    public long createLivingWill(LivingWillVO livingWillVO) throws DAOException {
        try {
            final String sql = "INSERT INTO LivingWill (Version, ValidFrom, ValidTo, ModifiedDateTime, PersonIdentifier, NoLifeProlongingIfDying, NoLifeProlongingIfSeverelyDegraded) " +
                    "VALUES (:Version, :ValidFrom, :ValidTo, :ModifiedDateTime, :PersonIdentifier, :NoLifeProlongingIfDying, :NoLifeProlongingIfSeverelyDegraded)";
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("Version", livingWillVO.getVersion(), Types.INTEGER)
                    .addValue("ValidFrom", java.sql.Timestamp.from(livingWillVO.getValidFrom().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("ValidTo", java.sql.Timestamp.from(livingWillVO.getValidTo().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("ModifiedDateTime", java.sql.Timestamp.from(livingWillVO.getModifiedDateTime().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("PersonIdentifier", livingWillVO.getPersonIdentifier(), Types.VARCHAR)
                    .addValue("NoLifeProlongingIfDying", livingWillVO.isNoLifeProlongingIfDying(), Types.TINYINT)
                    .addValue("NoLifeProlongingIfSeverelyDegraded", livingWillVO.isNoLifeProlongingIfSeverelyDegraded(), Types.TINYINT);
            final GeneratedKeyHolder generatedKeyHolder = new GeneratedKeyHolder();
            namedParamJdbcTemplate.update(sql, parameterSource, generatedKeyHolder);
            final long pid = Objects.requireNonNull(generatedKeyHolder.getKey()).longValue();
            livingWillVO.setLivingWillPID(pid);
            return pid;
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred inserting LivingWill", e);
        }
    }

    @Override
    public void updateValidToTimestamp(long pid, Instant newValidTo) throws DAOException {
        try {
            final String sql = "UPDATE LivingWill SET ValidTo = :ValidTo " +
                    "WHERE LivingWillPID = :LivingWillPID";
            final MapSqlParameterSource parameters = new MapSqlParameterSource()
                    .addValue("ValidTo", java.sql.Timestamp.from(newValidTo.truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("LivingWillPID", pid, Types.INTEGER);
            namedParamJdbcTemplate.update(sql, parameters);
        } catch (DataAccessException e) {
            throw new DAOException("An error updating valid to timestamp", e);
        }
    }

    @Override
    public void deleteLivingWillRecords(String personIdentifier) throws DAOException {
        try {
            String sql = "DELETE FROM LivingWill WHERE PersonIdentifier = :PersonIdentifier";
            final MapSqlParameterSource parameters = new MapSqlParameterSource()
                    .addValue("PersonIdentifier", personIdentifier, Types.VARCHAR);
            namedParamJdbcTemplate.update(sql, parameters);
        } catch (DataAccessException e) {
            throw new DAOException("An error occurred deleting from LivingWill", e);
        }
    }

}
