/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.minlog.registration.MinLogRegistrationService;
import dk.nsi.minlog.service.LogDataEntry;
import dk.nsi.minlog.service.MinLogRegistration;
import dk.nsi.minlog.service.Organization;
import dk.nsi.minlog.service.Organization.OrganizationTypeEnum;
import dk.nsi.service.ClientConfig;
import dk.nsi.service.DGWSHeaderWrapper;
import dk.nsi.util.Config;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.OrganizationTypeVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.SlaLogger;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.core.io.ClassPathResource;

import javax.xml.ws.Holder;
import java.net.MalformedURLException;
import java.net.URL;

public class MinLogClient {

    private static final Logger log = LogManager.getLogger(MinLogClient.class);

    private MinLogRegistration port;
    private MinLogRegistrationService service;
    private SlaLogger slaLogger = new SlaLogger("MinLog");

    public MinLogClient(String configName, String configFileName) throws MinLogException {
        Config config;
        String wsdlLocation;
        try {
            log.debug("Loading Minlog configuration from " + configFileName);
            config = new ClientConfig(configName, new ClassPathResource(configFileName).getFile());
            wsdlLocation = config.getProperty(MinLogRegistrationService.CONFIG_KEY);
        } catch (Exception e) {
            throw new MinLogException("Error reading MinLog configuration from: " + configFileName, e);
        }

        URL wsdl;
        try {
            wsdl = new URL(wsdlLocation);
        } catch (MalformedURLException e) {
            throw new MinLogException("MinLog configuration error. Malformed URL: " + wsdlLocation, e);
        }

        try {
            service = new MinLogRegistrationService(config, wsdl);
            log.debug("Starting MinLog registration service using wdsl from " + wsdl.toString());
            service.initialise();
            port = service.getMinLogRegistrationPort();
        } catch (javax.xml.ws.WebServiceException e) {
            throw new MinLogException("Could not retrieve MinLog wsdl from " + wsdl.toString(), e);
        } catch (Exception e) {
            throw new MinLogException("Could not initialize MinLog client", e);
        }
    }

    void logDataAdd(RequestContextVO context, MinLogData minLogData) throws MinLogException {
        DGWSHeaderWrapper signedHeaders;
        Holder<Header> medcomHeader;
        LogDataEntry minlogEntry;

        try {
            throwIfNull(minLogData.getCitizenCpr(), "CPR number must be specified for the person whose data is being viewed/changed.");
            throwIfNull(minLogData.getActivity(), "The activity being performed must be specified");
            throwIfNull(minLogData.getEventDateTime(), "Event time must be specified");
            throwIfNull(minLogData.getSourceSystemIdentifier(), "The source system must be specified");
            throwIfNull(minLogData.getUserCPR(), "The user CPR must be specified");
        } catch (IllegalArgumentException e) {
            throw new MinLogException("Invalid argument when calling the MinLog service", e);
        }

        try {
            signedHeaders = service.getSignedHeaders("0");
            medcomHeader = new Holder<>(signedHeaders.getMedcomHeader());
            minlogEntry = createLogDataEntry(minLogData);
        } catch (Exception e) {
            throw new MinLogException("Error creating MinLog data.", e);
        }

        long startMillis = System.currentTimeMillis();
        long startNanos = System.nanoTime();
        try {
            port.logDataAdd(minlogEntry, medcomHeader, signedHeaders.getSecurityHeader());
            addSlaLog(context, startNanos, startMillis, null);
        } catch (Exception e) {
            String errorMessage = "Error calling MinLog service";
            String msg = e.getMessage();
            if (msg != null) {
                errorMessage += ": " + msg;
            }
            addSlaLog(context, startNanos, startMillis, errorMessage);
            throw new MinLogException(errorMessage, e);
        }
    }

    private LogDataEntry createLogDataEntry(MinLogData minLogData) {
        final LogDataEntry minlogEntry;
        minlogEntry = new LogDataEntry();

        minlogEntry.setPersonCivilRegistrationIdentifier(minLogData.getCitizenCpr());
        minlogEntry.setActivity(minLogData.getActivity());
        minlogEntry.setEventDateTime(minLogData.getEventDateTime());
        minlogEntry.setSourceSystemIdentifier(minLogData.getSourceSystemIdentifier());
        minlogEntry.setUserIdentifier(minLogData.getUserCPR());

        if (minLogData.getOrganizationId() != null) {
            minlogEntry.setHealthcareProfessionalOrganization(
                    new Organization(minLogData.getOrganizationId(),
                            getMinlogOrganizationType(minLogData.getOrganizationType())));
        }

        if (minLogData.getOrganizationName() != null) {
            minlogEntry.setHealthcareProfessionalOrganizationName(minLogData.getOrganizationName());
        }

        String sessionId = minLogData.getSessionId();
        if (sessionId != null) {
            if (sessionId.length() > 46) {
                // MinLog accepts SessionIds of max 46 chars. Use sha1 if too long.
                sessionId = DigestUtils.sha1Hex(sessionId);
            }
            minlogEntry.setSessionId(sessionId);
        }

        if (minLogData.getOnBehalfOfCpr() != null) {
            minlogEntry.setUserIdentifierOnBehalfOf(minLogData.getOnBehalfOfCpr());
        }
        return minlogEntry;
    }

    private static OrganizationTypeEnum getMinlogOrganizationType(OrganizationTypeVO orgType) {
        switch (orgType) {
            case COMMUNALNUMBER:
                return OrganizationTypeEnum.COMMUNALNUMBER;
            case CVRNUMBER:
                return OrganizationTypeEnum.CVRNUMBER;
            case PNUMBER:
                return OrganizationTypeEnum.PNUMBER;
            case SKS:
                return OrganizationTypeEnum.SHAK;
            case SOR:
                return OrganizationTypeEnum.SOR;
            case YDERNUMMER:
                return OrganizationTypeEnum.YDERNUMMER;
            default:
                throw new RuntimeException("Unexpected organization type " + orgType);
        }
    }

    private <T> void throwIfNull(T value, String errorMessage) throws IllegalArgumentException {
        if (value == null) {
            throw new IllegalArgumentException(errorMessage);
        }
    }

    private void addSlaLog(RequestContextVO context, long startNanos, long startMillis, String errorMessage) {
        if (slaLogger != null) {
            slaLogger.addSlaLog(context, startNanos, startMillis, errorMessage);
        }
    }
}
