/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.util;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;

@Component
public class Properties {

    private static final Logger logger = LogManager.getLogger(Properties.class);

    @Value("${spring.application.name:}")
    private String applicationName;

    @Value("${datasource.ltr.jndi-name:}")
    private String dataSourceLtrJNDIName;

    @Value("${datasource.btr.jndi-name:}")
    private String dataSourceBtrJNDIName;

    @Value("${datasource.stamdata.jndi-name:}")
    private String dataSourceStamdataJNDIName;

    @Value("${dcc.endpoint.ltr:}")
    private String dccLTREndpoint;

    @Value("${dcc.endpoint.btr:}")
    private String dccBTREndpoint;

    @Value("${minlog.read-activity-text.ltr:}")
    private String minLogReadActivityTextLTR;

    @Value("${minlog.read-activity-text.btr:}")
    private String minLogReadActivityTextBTR;

    @Value("${schemavalidation.validate-requests:}")
    private boolean schemavalidationValidateRequests;

    @Value("${schemavalidation.validate-responses:}")
    private boolean schemavalidationValidateResponse;

    @Value("${schemavalidation.fail-on-response-error:}")
    private boolean schemavalidationFailOnResponseError;

    @Value("${forward-only-filter.enabled:}")
    private boolean enableForwardOnlyFilter;

    @Value("${ltr.close-date:}")
    private String ltrCloseDateString;
    private ZonedDateTime ltrCloseDate;

    @Value("${btr.open-date:}")
    private String btrOpenDateString;
    private ZonedDateTime btrOpenDate;
    
    // Delete job LTR
    @Value("${jobs.delete.enabled.ltr:}")
    private boolean jobsDeleteLtrEnabled;
    
    @Value("${jobs.delete.schedule.ltr:}")
    private String jobsDeleteLtrSchedule;
    
    @Value("${jobs.delete.cpr-max-results.ltr:}")
    private int jobsDeleteLtrCprMaxResults;
    
    @Value("${jobs.delete.cpr-max-loops.ltr:}")
    private int jobsDeleteLtrCprMaxLoops;

    // Delete job BTR
    @Value("${jobs.delete.enabled.btr:}")
    private boolean jobsDeleteBtrEnabled;
    
    @Value("${jobs.delete.schedule.btr:}")
    private String jobsDeleteBtrSchedule;
    
    @Value("${jobs.delete.cpr-max-results.btr:}")
    private int jobsDeleteBtrCprMaxResults;
    
    @Value("${jobs.delete.cpr-max-loops.btr:}")
    private int jobsDeleteBtrCprMaxLoops;

    @PostConstruct
    private void init() {
        try {
            ltrCloseDate = ZonedDateTime.parse(ltrCloseDateString);
            logger.info("LTR close date set to: " + getLtrCloseDate().format(DateTimeFormatter.ISO_DATE_TIME));
        } catch (DateTimeParseException e) {
            logger.warn("The date in the property 'ltr.close-date' could not be parsed. LTR registrations can be created.");
        }

        try {
            btrOpenDate = ZonedDateTime.parse(btrOpenDateString);
            logger.info("BTR open date set to: " + getBtrOpenDate().format(DateTimeFormatter.ISO_DATE_TIME));
        } catch (DateTimeParseException e) {
            logger.warn("The date in the property 'btr.open-date' could not be parsed. BTR registrations can be created.");
        }

        ForwardOnlyFilter.setEnableForwardOnlyFilter(isEnableForwardOnlyFilter());

        logger.info(toString());
    }

    @Override
    public String toString() {
        return "Configuration properties: {" +
                "applicationName=" + getApplicationName() +
                ", dataSourceLtrJNDIName=" + getDataSourceLtrJNDIName() +
                ", dataSourceBtrJNDIName=" + getDataSourceBtrJNDIName() +
                ", dataSourceStamdataJNDIName=" + getDataSourceStamdataJNDIName() +
                ", dccLTREndpoint=" + getDccLTREndpoint() +
                ", dccBTREndpoint=" + getDccBTREndpoint() +
                ", minLogReadActivityTextLTR=" + getMinLogReadActivityTextLTR() +
                ", minLogReadActivityTextBTR=" + getMinLogReadActivityTextBTR() +
                ", schemavalidationValidateRequests=" + isSchemavalidationValidateRequests() +
                ", schemavalidationValidateResponse=" + isSchemavalidationValidateResponse() +
                ", schemavalidationFailOnResponseError=" + isSchemavalidationFailOnResponseError() +
                ", enableForwardOnlyFilter=" + isEnableForwardOnlyFilter() +
                ", ltrCloseDate=" + getLtrCloseDate() +
                ", btrOpenDate=" + getBtrOpenDate() +
                ", jobsDeleteLtrEnabled=" + isJobsDeleteLtrEnabled() +
                ", jobsDeleteLtrSchedule=" + getJobsDeleteLtrSchedule() +
                ", jobsDeleteLtrCprMaxResults=" + getJobsDeleteLtrCprMaxResults() +
                ", jobsDeleteLtrCprMaxLoops=" + getJobsDeleteLtrCprMaxLoops() +
                ", jobsDeleteBtrEnabled=" + isJobsDeleteBtrEnabled() +
                ", jobsDeleteBtrSchedule=" + getJobsDeleteBtrSchedule() +
                ", jobsDeleteBtrCprMaxResults=" + getJobsDeleteBtrCprMaxResults() +
                ", jobsDeleteBtrCprMaxLoops=" + getJobsDeleteBtrCprMaxLoops() +
                '}';
    }

    public String getApplicationName() {
        return applicationName;
    }

    public String getDataSourceLtrJNDIName() {
        return dataSourceLtrJNDIName;
    }

    public String getDataSourceBtrJNDIName() {
        return dataSourceBtrJNDIName;
    }

    public String getDataSourceStamdataJNDIName() {
        return dataSourceStamdataJNDIName;
    }

    public String getDccLTREndpoint() {
        return dccLTREndpoint;
    }

    public String getDccBTREndpoint() {
        return dccBTREndpoint;
    }

    public String getMinLogReadActivityTextLTR() {
        return minLogReadActivityTextLTR;
    }

    public String getMinLogReadActivityTextBTR() {
        return minLogReadActivityTextBTR;
    }

    public boolean isSchemavalidationValidateRequests() {
        return schemavalidationValidateRequests;
    }

    public boolean isSchemavalidationValidateResponse() {
        return schemavalidationValidateResponse;
    }

    public boolean isSchemavalidationFailOnResponseError() {
        return schemavalidationFailOnResponseError;
    }

    public boolean isEnableForwardOnlyFilter() {
        return enableForwardOnlyFilter;
    }

    public ZonedDateTime getLtrCloseDate() {
        return ltrCloseDate;
    }

    public ZonedDateTime getBtrOpenDate() {
        return btrOpenDate;
    }

    public boolean isJobsDeleteLtrEnabled() {
        return jobsDeleteLtrEnabled;
    }

    public String getJobsDeleteLtrSchedule() {
        return jobsDeleteLtrSchedule;
    }

    public int getJobsDeleteLtrCprMaxResults() {
        return jobsDeleteLtrCprMaxResults;
    }

    public int getJobsDeleteLtrCprMaxLoops() {
        return jobsDeleteLtrCprMaxLoops;
    }

    public boolean isJobsDeleteBtrEnabled() {
        return jobsDeleteBtrEnabled;
    }

    public String getJobsDeleteBtrSchedule() {
        return jobsDeleteBtrSchedule;
    }

    public int getJobsDeleteBtrCprMaxResults() {
        return jobsDeleteBtrCprMaxResults;
    }

    public int getJobsDeleteBtrCprMaxLoops() {
        return jobsDeleteBtrCprMaxLoops;
    }

}
