/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.LivingWillService;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.LivingWillServiceImpl;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.proxydata.ProxyMetaData;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.ws.server.endpoint.annotation.RequestPayload;

@Component
public class LivingWillWebServiceImpl extends AbstractWebService implements LivingWillWebService {
    private final boolean isAdmin = false;

    private static AuditLogger auditLogger = new AuditLogger(LivingWillServiceImpl.class);

    private static final String NAMESPACE_URI = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01/";

    private final LivingWillService businessService;
    private final LivingWillHL7CDAMapper livingWillHL7CDAMapper;
    private ProxyMetaData proxyMetaData;

    @Autowired
    public LivingWillWebServiceImpl(LivingWillService businessService, LivingWillHL7CDAMapper livingWillHL7CDAMapper, ProxyMetaData proxyMetaData) {
        this.businessService = businessService;
        this.livingWillHL7CDAMapper = livingWillHL7CDAMapper;
        this.proxyMetaData = proxyMetaData;
    }

    @Override
    public GetLivingWillResponse getLivingWill(@RequestPayload GetLivingWillRequest request, boolean isAdmin) throws Exception {
        final GetLivingWillResponse response = new GetLivingWillResponse();

        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new LivingWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);

            final LivingWillVO livingWillVO = businessService.getLivingWill(metaData.getActor(), metaData.getContext(), metaData.getProxyData());

            final POCDMT000040ClinicalDocument clinicalDocument = livingWillHL7CDAMapper.toClinicalDocument(livingWillVO, metaData.getContext());
            response.setClinicalDocument(clinicalDocument);
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.GET_LIVING_WILL_VALIDATION_ERROR, FaultCode.GET_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public HasLivingWillResponse hasLivingWill(HasLivingWillRequest request, boolean isAdmin) throws Exception {
        final HasLivingWillResponse response = new HasLivingWillResponse();

        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new LivingWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);

            final boolean hasLivingWill = businessService.hasLivingWill(metaData.getContext(), metaData.getProxyData());

            response.setWillExists(hasLivingWill);
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.HAS_LIVING_WILL_VALIDATION_ERROR, FaultCode.HAS_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public CreateLivingWillResponse createLivingWill(CreateLivingWillRequest request, boolean isAdmin) throws Exception {
        final CreateLivingWillResponse response = new ObjectFactory().createCreateLivingWillResponse();

        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new LivingWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);
            final LivingWillVO livingWillVO = livingWillHL7CDAMapper.createLivingWillFromRequest(
                    metaData.getContext().getPersonIdentifier(),
                    request.getLivingWill()
            );
            businessService.createLivingWill(livingWillVO, metaData.getActor(), metaData.getContext(), metaData.getProxyData());
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.CREATE_LIVING_WILL_VALIDATION_ERROR, FaultCode.CREATE_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public UpdateLivingWillResponse updateLivingWill(UpdateLivingWillRequest request, boolean isAdmin) throws Exception {
        final UpdateLivingWillResponse response = new ObjectFactory().createUpdateLivingWillResponse();

        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new LivingWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);
            final LivingWillVO livingWillVO = livingWillHL7CDAMapper.createLivingWillFromRequest(
                    metaData.getContext().getPersonIdentifier(),
                    request.getLivingWill()
            );
            businessService.updateLivingWill(livingWillVO, metaData.getActor(), metaData.getContext(), metaData.getProxyData());
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.UPDATE_LIVING_WILL_VALIDATION_ERROR, FaultCode.UPDATE_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public DeleteLivingWillResponse deleteLivingWill(DeleteLivingWillRequest request, boolean isAdmin) throws Exception {
        final DeleteLivingWillResponse response = new ObjectFactory().createDeleteLivingWillResponse();

        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new LivingWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);
            businessService.deleteLivingWill(metaData.getActor(), metaData.getContext(), metaData.getProxyData());
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.DELETE_LIVING_WILL_VALIDATION_ERROR, FaultCode.DELETE_LIVING_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

}
