/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillService;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillServiceImpl;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.proxydata.ProxyMetaData;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class TreatmentWillWebServiceImpl extends AbstractWebService implements TreatmentWillWebService {
    private static AuditLogger auditLogger = new AuditLogger(TreatmentWillServiceImpl.class);

    private final TreatmentWillService businessService;
    private final TreatmentWillHL7CDAMapper treatmentWillHL7CDAMapper;
    private ProxyMetaData proxyMetaData;

    @Autowired
    public TreatmentWillWebServiceImpl(TreatmentWillService businessService, TreatmentWillHL7CDAMapper treatmentWillHL7CDAMapper, ProxyMetaData proxyMetaData) {
        this.businessService = businessService;
        this.treatmentWillHL7CDAMapper = treatmentWillHL7CDAMapper;
        this.proxyMetaData = proxyMetaData;
    }

    @Override
    public GetTreatmentWillResponse getTreatmentWill(GetTreatmentWillRequest request, boolean isAdmin) throws Exception {
        final GetTreatmentWillResponse response = new GetTreatmentWillResponse();
        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new TreatmentWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);

            final TreatmentWillVO treatmentWillVO = businessService.getTreatmentWill(metaData.getActor(), metaData.getContext(), metaData.getProxyData());

            final POCDMT000040ClinicalDocument clinicalDocument = treatmentWillHL7CDAMapper.toClinicalDocument(treatmentWillVO, metaData.getContext());
            response.setClinicalDocument(clinicalDocument);
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.GET_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public HasTreatmentWillResponse hasTreatmentWill(HasTreatmentWillRequest request, boolean isAdmin) throws Exception {
        final HasTreatmentWillResponse response = new HasTreatmentWillResponse();
        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new TreatmentWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);

            final boolean hasTreatmentWill = businessService.hasTreatmentWill(metaData.getContext(), metaData.getProxyData());

            response.setWillExists(hasTreatmentWill);
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.HAS_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.HAS_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public CreateTreatmentWillResponse createTreatmentWill(CreateTreatmentWillRequest request, boolean isAdmin) throws Exception {
        final CreateTreatmentWillResponse response = new CreateTreatmentWillResponse();

        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new TreatmentWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);
            final TreatmentWillVO treatmentWillVO = treatmentWillHL7CDAMapper.fromTreatmentWill(
                    metaData.getContext().getPersonIdentifier(),
                    request.getTreatmentWill()
            );
            businessService.createTreatmentWill(treatmentWillVO, metaData.getActor(), metaData.getContext(), metaData.getProxyData(), false);
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.CREATE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.CREATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public UpdateTreatmentWillResponse updateTreatmentWill(UpdateTreatmentWillRequest request, boolean isAdmin) throws Exception {
        final UpdateTreatmentWillResponse response = new UpdateTreatmentWillResponse();

        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new TreatmentWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);
            final TreatmentWillVO treatmentWillVO = treatmentWillHL7CDAMapper.fromTreatmentWill(
                    metaData.getContext().getPersonIdentifier(),
                    request.getTreatmentWill()
            );
            businessService.updateTreatmentWill(treatmentWillVO, metaData.getActor(), metaData.getContext(), metaData.getProxyData(), false);
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.UPDATE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.UPDATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }

    @Override
    public DeleteTreatmentWillResponse deleteTreatmentWill(DeleteTreatmentWillRequest request, boolean isAdmin) throws Exception {
        final DeleteTreatmentWillResponse response = new DeleteTreatmentWillResponse();

        ProxyMetaData.RequestMetaData metaData = null;
        try {
            new TreatmentWillHL7CDAValidator().validate(request);
            metaData = proxyMetaData.getRequestMetadata(request.getId().getExtension(), isAdmin);
            businessService.deleteTreatmentWill(metaData.getActor(), metaData.getContext(), metaData.getProxyData());
        } catch (Exception e) {
            auditLogger.auditError(metaData, e);
            commonExceptionHandler(e, FaultCode.DELETE_TREATMENT_WILL_VALIDATION_ERROR, FaultCode.DELETE_TREATMENT_WILL_INTERNAL_SERVER_ERROR);
        }

        return response;
    }
}
