/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice.proxydata;

import dk.sds.dgws_idws_proxy.security.ProxyData;
import dk.sds.dgws_idws_proxy.security.RequestType;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sdsd.nsp.slalog.ws.SLALoggingServletFilter;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.OrganizationTypeVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.UserType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.servlet.http.HttpServletRequest;
import java.time.Instant;

@Component
public class ProxyMetaData {
    static final int LEVEL_FOCES = 3;
    static final int LEVEL_MOCES = 4;

    private final ProxyDataProvider proxyDataProvider;
    private final HttpServletRequest httpServletRequest;

    @Autowired
    public ProxyMetaData(ProxyDataProvider proxyDataProvider, HttpServletRequest httpServletRequest) {
        this.proxyDataProvider = proxyDataProvider;
        this.httpServletRequest = httpServletRequest;
    }

    public RequestMetaData getRequestMetadata(String personIdentifier, boolean isAdmin) {
        ProxyData headerInfo = proxyDataProvider.getProxyData();
        final RequestMetaData metaData = new RequestMetaData();
        metaData.setActor(getServiceActor(headerInfo));
        metaData.setContext(getRequestContext(headerInfo, personIdentifier, isAdmin));
        metaData.setProxyData(headerInfo);
        return metaData;
    }

    private ServiceActorVO getServiceActor(ProxyData headerInfo) {
        ServiceActorVO actor = new ServiceActorVO();
        actor.setOrganizationId(headerInfo.getCareProviderID());
        actor.setOrganizationIdSource(OrganizationTypeVO.fromMedcomType(headerInfo.getCareProviderType()).name());
        actor.setOrganizationName(headerInfo.getCareProviderName());
        actor.setPersonIdentifier(headerInfo.getCprnr());

        addHeaderInfoToActor(actor, headerInfo.getAuthorizationCode(), headerInfo.getOccupation());

        return actor;
    }

    private static void addHeaderInfoToActor(ServiceActorVO actor, String authorizationCode, String occupation) {
        if (authorizationCode != null) {
            actor.setAuthorisationNumber(authorizationCode.trim());
        }

        if (occupation != null) {
            actor.setEducationCode(occupation.trim());
        }
    }

    private RequestContextVO getRequestContext(ProxyData headerInfo, String personIdentifier, boolean isAdmin) {
        RequestContextVO context = new RequestContextVO();
        context.setTime(Instant.now()); // Get current time for request
        context.setPersonIdentifier(personIdentifier);
        context.setFlowId(headerInfo.getFlowId());
        context.setMessageId(headerInfo.getMessageId());
        context.setSystemIdentifier(headerInfo.getItSystemName());

        SLALogItem slaLogItem = (SLALogItem) httpServletRequest.getAttribute(SLALoggingServletFilter.CURRENT_LOGITEM_REQUEST_KEY);
        if (slaLogItem != null) {
            slaLogItem.setMessageId(headerInfo.getMessageId());
        }
        context.setSlaLogItem(slaLogItem);

        UserType userType = findUserType(headerInfo, personIdentifier, isAdmin);
        context.setUserType(userType);

        return context;
    }

    UserType findUserType(ProxyData proxyData, String personIdentifier, boolean isAdmin) {
        UserType result;

        int authLevelOrNull = -1;
        try {
            authLevelOrNull = Integer.parseInt(proxyData.getAuthenticationLevel());
        } catch (NumberFormatException e) {
            // Ignore
        }

        if (isAdmin && proxyData.getRequestType() != null && proxyData.getRequestType().equals(RequestType.DGWS) && authLevelOrNull == LEVEL_MOCES) {
            result = UserType.Admin;
        } else if (proxyData.getRequestType() != null && proxyData.getRequestType().equals(RequestType.DGWS) && authLevelOrNull == LEVEL_MOCES) {
            result = UserType.HealthCareProfessional;
        } else if (proxyData.getRequestType() != null && proxyData.getRequestType().equals(RequestType.DGWS) && authLevelOrNull == LEVEL_FOCES) {
            result = UserType.SystemUser;
        } else if (proxyData.getRequestType() != null && proxyData.getRequestType().equals(RequestType.IDWS) && proxyData.getCprnr() != null && personIdentifier.trim().equals(proxyData.getCprnr().trim())) {
            result = UserType.DataOwner;
        } else {
            result = UserType.Other;
        }

        return result;
    }

    public class RequestMetaData {
        private ServiceActorVO actor;
        private RequestContextVO context;
        private ProxyData proxyData;

        public ServiceActorVO getActor() {
            return actor;
        }

        public void setActor(ServiceActorVO actor) {
            this.actor = actor;
        }

        public RequestContextVO getContext() {
            return context;
        }

        public void setContext(RequestContextVO context) {
            this.context = context;
        }

        public void setProxyData(ProxyData headerInfo) {
            this.proxyData = headerInfo;
        }

        public ProxyData getProxyData() {
            return proxyData;
        }
    }

}
