/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import dk.sundhedsdatastyrelsen.livstestamente.service.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.UserType;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

public class AbstractWillAccessControlTests extends TestHelper {

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    private boolean isOpen;
    private AbstractWillAccessControlImpl classUnderTest;

    @Before
    public void setUp() {
        isOpen = true;
        classUnderTest = new AbstractWillAccessControlImpl() {
            @Override
            boolean isOpen(RequestContextVO context) {
                return isOpen;
            }
        };
    }

    @Test
    public void checkGetAccess_success() {
        // Arrange
        RequestContextVO requestContext1 = getRequestContextExample();
        RequestContextVO requestContext2 = getRequestContextExample();
        requestContext2.setUserType(UserType.HealthCareProfessional);

        // Act
        classUnderTest.checkGetAccess(requestContext1);
        classUnderTest.checkGetAccess(requestContext2);
    }

    @Test
    public void checkGetAccess_fail_wrongUserType() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.SystemUser);

        // Act
        try {
            classUnderTest.checkGetAccess(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED, e.getFaultCode());
        }
    }

    @Test
    public void checkHasAccess_success() {
        // Arrange
        RequestContextVO requestContext1 = getRequestContextExample();
        RequestContextVO requestContext2 = getRequestContextExample();
        requestContext2.setUserType(UserType.SystemUser);

        // Act
        classUnderTest.checkHasAccess(requestContext1);
        classUnderTest.checkHasAccess(requestContext2);
    }

    @Test
    public void checkHasAccess_fail_wrongUserType() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.Other);

        // Act
        try {
            classUnderTest.checkHasAccess(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED, e.getFaultCode());
        }
    }

    @Test
    public void checkCreateAccess_success() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();

        // Act
        classUnderTest.checkCreateAccess(requestContext);
    }

    @Test
    public void checkCreateAccess_fail_wrongUserType() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.HealthCareProfessional);

        // Act
        try {
            classUnderTest.checkCreateAccess(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED, e.getFaultCode());
        }
    }

    @Test
    public void checkCreateAccess_fail_notWriteable() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();
        isOpen = false;

        // Act
        try {
            classUnderTest.checkCreateAccess(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.CREATE_NOT_ALLOWED, e.getFaultCode());
        }
    }

    @Test
    public void checkUpdateAccess_success() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();

        // Act
        classUnderTest.checkUpdateAccess(requestContext);
    }

    @Test
    public void checkUpdateAccess_fail_wrongUserType() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.HealthCareProfessional);

        // Act
        try {
            classUnderTest.checkUpdateAccess(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED, e.getFaultCode());
        }
    }

    @Test
    public void checkDeleteAccess_success() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();

        // Act
        classUnderTest.checkDeleteAccess(requestContext);
    }

    @Test
    public void checkDeleteAccess_fail_wrongUserType() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.HealthCareProfessional);

        // Act
        try {
            classUnderTest.checkDeleteAccess(requestContext);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED, e.getFaultCode());
        }
    }
}
