/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import dk.cprsubscriber.dao.impl.CPRSubscriberTest;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.PropertiesDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.PropertiesVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.transaction.annotation.Transactional;

import javax.sql.DataSource;
import java.time.Instant;
import java.time.Year;
import java.time.ZoneId;
import java.time.ZonedDateTime;

import static dk.sundhedsdatastyrelsen.livstestamente.service.business.DeleteTreatmentWillJob.LAST_DELETION_TIME;
import static dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillServiceImpl.VALID_TO_DEFAULT;
import static org.junit.Assert.*;

@RunWith(SpringRunner.class)
@SpringBootTest
@Transactional("stamdataDataSourceTxManager")
public class DeleteTreatmentWillJobTest extends CPRSubscriberTest {

    @Autowired
    DeleteTreatmentWillJob deleteTreatmentWillJob;

    @Autowired
    private DataSource stamdataDataSource;

    @Autowired
    TreatmentWillDAO treatmentWillDAO;

    @Autowired
    @Qualifier("treatmentWillPropertiesDAOImpl")
    private PropertiesDAO propertiesDAO;

    @Autowired
    private TestHelper testHelper;

    private static boolean initialized = false;

    final private String DEFAULT_PERSONID = "1111111111";
    private Instant now;

    @Before
    public void setup() throws DAOException {
        this.namedParamJdbcTemplate = new NamedParameterJdbcTemplate(stamdataDataSource);

        now = Instant.now();
        propertiesDAO.updateProperty(LAST_DELETION_TIME, now.minusSeconds(60).toString());
    }

    @Test
    public void deletebtrJob_PersonIsDeadOneYearAgo_btrDeleted() throws DAOException, InterruptedException {
        String lastDeletionTime = now.minusSeconds(60).toString();
        propertiesDAO.updateProperty(LAST_DELETION_TIME, lastDeletionTime);

        createAlivePeople();
        createTreatmentWill(now);
        runDeletionJob();

        TreatmentWillVO btrBeforeDeletion = treatmentWillDAO.getCurrentTreatmentWill(DEFAULT_PERSONID, now);
        assertNotNull(btrBeforeDeletion);
        assertNotNull(btrBeforeDeletion.getTreatmentWillPID());
        assertEquals(DEFAULT_PERSONID, btrBeforeDeletion.getPersonIdentifier());

        final Instant date29991231 = Year.of(2999).atMonth(12).atDay(31).atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
        registerDeathOneYearAgo(date29991231);
        runDeletionJob();
        boolean btrAfterDeletion = treatmentWillDAO.existsTreatmentWillRecords(DEFAULT_PERSONID);
        assertFalse(btrAfterDeletion);

        // Assert that correct property table was updated
        PropertiesVO lastDeletionTimeUpdated = propertiesDAO.getProperty(LAST_DELETION_TIME);
        assertNotEquals(lastDeletionTime, lastDeletionTimeUpdated.getPropertyValue());
    }

    @Test
    public void deletebtrJob_PersonIsDeadOneYearAgoCurrentlyAlive_btrNotDeleted() throws DAOException, InterruptedException {
        String lastDeletionTime = now.minusSeconds(60).toString();
        propertiesDAO.updateProperty(LAST_DELETION_TIME, lastDeletionTime);

        createAlivePeople();
        createTreatmentWill(now);
        runDeletionJob();

        TreatmentWillVO btrBeforeDeletion = treatmentWillDAO.getCurrentTreatmentWill(DEFAULT_PERSONID, now);
        assertNotNull(btrBeforeDeletion);
        assertNotNull(btrBeforeDeletion.getTreatmentWillPID());
        assertEquals(DEFAULT_PERSONID, btrBeforeDeletion.getPersonIdentifier());

        registerDeathOneYearAgo(getDateTwoMonthsAgo());
        registerPersonAlive();
        runDeletionJob();

        TreatmentWillVO btrAfterDeletion = treatmentWillDAO.getCurrentTreatmentWill(DEFAULT_PERSONID, now);
        assertNotNull(btrAfterDeletion);
        assertNotNull(btrAfterDeletion.getTreatmentWillPID());
        assertEquals(DEFAULT_PERSONID, btrAfterDeletion.getPersonIdentifier());
    }

    private void runDeletionJob() throws InterruptedException {
        Thread.sleep(1000);
        deleteTreatmentWillJob.runDeleteTreatmentWill();
    }

    private void createAlivePeople() {
    	final Instant date29991231 = Year.of(2999).atMonth(12).atDay(31).atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
    	final Instant date20011224 = Year.of(2001).atMonth(12).atDay(24).atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
    	final Instant date19750505 = Year.of(1975).atMonth(5).atDay(5).atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
    	final Instant date19350101 = Year.of(1935).atMonth(1).atDay(1).atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
        
        createPerson(DEFAULT_PERSONID, "01", date19350101, Instant.now(), date29991231);
        createPerson("2222222222", "01", date19750505, Instant.now(), date29991231);
        createPerson("3333333333", "01", date20011224, Instant.now(), date29991231);
    	
    }

    private void registerDeathOneYearAgo(Instant validTo) {
    	final Instant date19350101 = Year.of(1935).atMonth(1).atDay(1).atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
    	closeCprRecord(DEFAULT_PERSONID, getDateOneYearAgo());
        createPerson(DEFAULT_PERSONID, "90", date19350101, getDateOneYearAgo(), validTo);
    }

    private void registerPersonAlive() {
    	final Instant date29991231 = Year.of(2999).atMonth(12).atDay(31).atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
    	final Instant date19350101 = Year.of(1935).atMonth(1).atDay(1).atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
    	createPerson(DEFAULT_PERSONID, "01", date19350101, Instant.now(), date29991231);
    }

    void createTreatmentWill(Instant now) throws DAOException {
        TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWillExample();
        treatmentWillVO.setPersonIdentifier(DEFAULT_PERSONID);
        treatmentWillVO.setValidFrom(now);
        treatmentWillVO.setModifiedDateTime(now);
        treatmentWillVO.setValidTo(VALID_TO_DEFAULT);
        treatmentWillVO.setVersion(1);
        treatmentWillDAO.createTreatmentWill(treatmentWillVO);
    }

    private Instant getDateOneYearAgo() {
    	ZonedDateTime zonedDateTime = ZonedDateTime.now();
        zonedDateTime = zonedDateTime.minusYears(1);
        return zonedDateTime.toInstant(); 
    }

    private Instant getDateTwoMonthsAgo() {
    	ZonedDateTime zonedDateTime = ZonedDateTime.now();
        zonedDateTime = zonedDateTime.minusMonths(2);
        return zonedDateTime.toInstant();        
    }

}
