/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.LivingWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.TreatmentWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.Properties;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;

import java.time.Instant;

import static org.junit.Assert.*;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.*;

public class LivingWillServiceTests extends TestHelper {

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    @Mock
    private Properties properties;

    @Mock
    private TreatmentWillDAO treatmentWillDAO;

    @Mock
    private LivingWillDAO livingWillDAO;

    @Mock
    private MinLogDataWrapper minLogDataWrapper;

    @Mock
    private AccessControl accessControl;

    @Mock
    private TreatmentWillValidator treatmentWillValidator;

    @Mock
    private LivingWillValidator livingWillValidator;

    private LivingWillServiceImpl classUnderTest;

    @Before
    public void setUp() {
        classUnderTest = new LivingWillServiceImpl(properties, treatmentWillDAO, livingWillDAO, minLogDataWrapper, accessControl, treatmentWillValidator, livingWillValidator);
    }

    @Test
    public void getLivingWill_all_DAO_fail() throws DAOException {
        // Note: These tests cover the thrown exceptions in the service,
        // but JaCoCo cannot detect lines that throw exceptions.

        // Arrange
        LivingWillVO livingWillVO = new LivingWillVO();
        ServiceActorVO serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample();
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to throw an exception
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenThrow(new DAOException("An error occurred reading LivingWill"));

        // Act
        try {
            classUnderTest.getLivingWill(serviceActor, requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.GET_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.hasLivingWill(requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.HAS_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.createLivingWill(livingWillVO, serviceActorExample, requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.CREATE_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.updateLivingWill(livingWillVO, serviceActorExample, requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.UPDATE_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.deleteLivingWill(serviceActorExample, requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.DELETE_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void getLivingWill_success() throws DAOException, MinLogException {
        // Arrange
        ServiceActorVO serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample();
        LivingWillVO livingWillVO = new LivingWillVO();

        // Stub DAO method to return the LivingWillVO
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVO);

        // Stub Properties to return a string
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogReadActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);

        // Act
        LivingWillVO livingWillReturned = classUnderTest.getLivingWill(serviceActor, requestContext, null);

        // Assert
        verify(accessControl).checkGetAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(livingWillValidator).validateLivingWillExists(eq(livingWillVO));
        verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER));

        assertEquals(livingWillVO, livingWillReturned);
    }

    @Test
    public void getLivingWill_failOnMinLogException() throws DAOException, MinLogException {
        // Arrange
        ServiceActorVO serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample();
        LivingWillVO livingWillVO = new LivingWillVO();

        // Stub DAO method to return the LivingWillVO
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVO);

        // Stub MinLog to throw an Exception
        doThrow(MinLogException.class).when(minLogDataWrapper)
                .logReadActivity(
                        any(RequestContextVO.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(OrganizationTypeVO.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class)
                );

        // Stub Properties to return a string
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogReadActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);

        // Act
        try {
            classUnderTest.getLivingWill(serviceActor, requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            verify(accessControl).checkGetAccess(eq(requestContext));
            verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
            verify(livingWillValidator).validateLivingWillExists(eq(livingWillVO));
            verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER));

            assertEquals(FaultCode.GET_LIVING_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void hasLivingWill_true() throws DAOException {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();
        LivingWillVO livingWillVO = new LivingWillVO();

        // Stub DAO method to return the LivingWillVO
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVO);

        // Act
        boolean hasLivingWill = classUnderTest.hasLivingWill(requestContext, null);

        // Assert
        verify(accessControl).checkHasAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));

        assertTrue(hasLivingWill);
    }

    @Test
    public void hasLivingWill_false() throws DAOException {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();

        // Stub DAO method to return null
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(null);

        // Act
        boolean hasLivingWill = classUnderTest.hasLivingWill(requestContext, null);

        // Assert
        verify(accessControl).checkHasAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));

        assertFalse(hasLivingWill);
    }

    @Test
    public void createLivingWill_success() throws DAOException {
        // Arrange
        LivingWillVO livingWillVO = getLivingWillExample();
        RequestContextVO requestContext = getRequestContextExample();
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return null (although the mock would do that anyway)
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class))).thenReturn(null);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class))).thenReturn(null);

        // Act
        classUnderTest.createLivingWill(livingWillVO, serviceActorExample, requestContext, null);

        // Assert
        verify(accessControl).checkCreateAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(livingWillValidator).validateLivingWillNotExists(isNull());
        verify(treatmentWillValidator).validateTreatmentWillNotExists(isNull());
        verify(livingWillDAO).createLivingWill(eq(livingWillVO));
        assertEquals(livingWillVO.getPersonIdentifier(), livingWillVO.getPersonIdentifier());
        assertEquals(LivingWillServiceImpl.INITIAL_VERSION_DEFAULT, livingWillVO.getVersion().intValue());
        assertEquals(requestContext.getTime(), livingWillVO.getValidFrom());
        assertEquals(LivingWillServiceImpl.VALID_TO_DEFAULT, livingWillVO.getValidTo());
        assertEquals(requestContext.getTime(), livingWillVO.getModifiedDateTime());
    }

    @Test
    public void adminCreateLivingWill_minLogCalled_success() throws DAOException, MinLogException {
        // Arrange
        LivingWillVO livingWillVO = getLivingWillExample();
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.Admin);
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return null (although the mock would do that anyway)
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class))).thenReturn(null);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class))).thenReturn(null);
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogReadActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);

        // Act
        classUnderTest.createLivingWill(livingWillVO, serviceActorExample, requestContext, null);

        // Assert
        verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER));
    }

    @Test
    public void updateLivingWill_success() throws DAOException {
        // Arrange
        long pid = 1L;
        int version = 1;

        LivingWillVO livingWillVO = getLivingWillExample();
        RequestContextVO requestContext = getRequestContextExample();
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return an object with a PID and Version
        LivingWillVO livingWillVOFake = new LivingWillVO();
        livingWillVOFake.setLivingWillPID(pid);
        livingWillVOFake.setVersion(version);
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVOFake);

        // Act
        classUnderTest.updateLivingWill(livingWillVO, serviceActorExample, requestContext, null);

        // Assert
        verify(accessControl).checkUpdateAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(livingWillValidator).validateLivingWillExists(eq(livingWillVOFake));
        verify(livingWillDAO).updateValidToTimestamp(eq(pid), eq(requestContext.getTime()));
        verify(livingWillDAO).createLivingWill(eq(livingWillVO));

        assertEquals(livingWillVO.getPersonIdentifier(), livingWillVO.getPersonIdentifier());
        assertEquals(LivingWillServiceImpl.INITIAL_VERSION_DEFAULT + 1, livingWillVO.getVersion().intValue());
        assertEquals(requestContext.getTime(), livingWillVO.getValidFrom());
        assertEquals(LivingWillServiceImpl.VALID_TO_DEFAULT, livingWillVO.getValidTo());
        assertEquals(requestContext.getTime(), livingWillVO.getModifiedDateTime());
    }

    @Test
    public void adminUpdateLivingWill__minLogCalled_success() throws DAOException, MinLogException {
        // Arrange
        long pid = 1L;
        int version = 1;

        LivingWillVO livingWillVO = getLivingWillExample();
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.Admin);
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return an object with a PID and Version
        LivingWillVO livingWillVOFake = new LivingWillVO();
        livingWillVOFake.setLivingWillPID(pid);
        livingWillVOFake.setVersion(version);
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVOFake);
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogReadActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);

        // Act
        classUnderTest.updateLivingWill(livingWillVO, serviceActorExample, requestContext, null);

        // Assert
        verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER));
    }

    @Test
    public void deleteLivingWill_success() throws DAOException {
        // Arrange
        long pid = 1L;
        int version = 1;

        RequestContextVO requestContext = getRequestContextExample();
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return an object with a PID and Version
        LivingWillVO livingWillVOFake = new LivingWillVO();
        livingWillVOFake.setLivingWillPID(pid);
        livingWillVOFake.setVersion(version);
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVOFake);

        // Act
        classUnderTest.deleteLivingWill(serviceActorExample, requestContext, null);

        // Assert
        verify(accessControl).checkDeleteAccess(eq(requestContext));

        ArgumentCaptor<Long> updateValidToArgument1 = ArgumentCaptor.forClass(Long.class);
        ArgumentCaptor<Instant> updateValidToArgument2 = ArgumentCaptor.forClass(Instant.class);
        verify(livingWillDAO).updateValidToTimestamp(updateValidToArgument1.capture(), updateValidToArgument2.capture());

        assertEquals(pid, updateValidToArgument1.getValue().longValue());
        assertNotNull(updateValidToArgument2.getValue());
    }

    @Test
    public void adminDeleteLivingWill_minLogCalled_success() throws DAOException, MinLogException {
        // Arrange
        long pid = 1L;
        int version = 1;

        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.Admin);
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return an object with a PID and Version
        LivingWillVO livingWillVOFake = new LivingWillVO();
        livingWillVOFake.setLivingWillPID(pid);
        livingWillVOFake.setVersion(version);
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class)))
                .thenReturn(livingWillVOFake);
        String minLogReadActivityTextLTR = "test";
        when(properties.getMinLogReadActivityTextLTR()).thenReturn(minLogReadActivityTextLTR);


        // Act
        classUnderTest.deleteLivingWill(serviceActorExample, requestContext, null);

        // Assert
        verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextLTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(LivingWillServiceImpl.SERVICE_IDENTIFIER));
    }
}
