/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.business;

import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.LivingWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.validator.TreatmentWillValidator;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.LivingWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.dao.TreatmentWillDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.Properties;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;

import java.time.Instant;

import static org.junit.Assert.*;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.*;

public class TreatmentWillServiceTests extends TestHelper {

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    @Mock
    private Properties properties;

    @Mock
    private TreatmentWillDAO treatmentWillDAO;

    @Mock
    private LivingWillDAO livingWillDAO;

    @Mock
    private MinLogDataWrapper minLogDataWrapper;

    @Mock
    private AccessControl accessControl;

    @Mock
    private TreatmentWillValidator treatmentWillValidator;

    @Mock
    private LivingWillValidator livingWillValidator;

    private TreatmentWillServiceImpl classUnderTest;

    @Before
    public void setUp() {
        classUnderTest = new TreatmentWillServiceImpl(properties, treatmentWillDAO, livingWillDAO, minLogDataWrapper, accessControl, treatmentWillValidator, livingWillValidator);
    }

    @Test
    public void getTreatmentWill_all_DAO_fail() throws DAOException {
        // Note: These tests cover the thrown exceptions in the service,
        // but JaCoCo cannot detect lines that throw exceptions.

        // Arrange
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();
        ServiceActorVO serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample();
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to throw an exception
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenThrow(new DAOException("An error occurred reading TreatmentWill"));

        // Act
        try {
            classUnderTest.getTreatmentWill(serviceActor, requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.hasTreatmentWill(requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.HAS_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.createTreatmentWill(treatmentWillVO, serviceActorExample, requestContext, null, false);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.CREATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.updateTreatmentWill(treatmentWillVO, serviceActorExample, requestContext, null, false);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.UPDATE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            classUnderTest.deleteTreatmentWill(serviceActorExample, requestContext, null);
            fail();
        } catch (ServiceException e) {
            // Assert
            assertEquals(FaultCode.DELETE_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void getTreatmentWill_success() throws DAOException, MinLogException {
        // Arrange
        ServiceActorVO serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample();
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();

        // Stub DAO method to return the TreatmentWillVO
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVO);

        // Stub Properties to return a string
        String minLogReadActivityTextBTR = "test";
        when(properties.getMinLogReadActivityTextBTR()).thenReturn(minLogReadActivityTextBTR);

        // Act
        TreatmentWillVO treatmentWillReturned = classUnderTest.getTreatmentWill(serviceActor, requestContext, null);

        // Assert
        verify(accessControl).checkGetAccess(eq(requestContext));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(treatmentWillValidator).validateTreatmentWillExists(eq(treatmentWillVO));
        verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER));

        assertEquals(treatmentWillVO, treatmentWillReturned);
    }

    @Test
    public void getTreatmentWill_failOnMinLogException() throws DAOException, MinLogException {
        // Arrange
        ServiceActorVO serviceActor = getServiceActorExample();
        RequestContextVO requestContext = getRequestContextExample();
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();

        // Stub DAO method to return the TreatmentWillVO
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVO);

        // Stub MinLog to throw an Exception
        doThrow(MinLogException.class).when(minLogDataWrapper)
                .logReadActivity(
                        any(RequestContextVO.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(OrganizationTypeVO.class),
                        nullable(String.class),
                        nullable(String.class),
                        nullable(String.class)
                );

        // Stub Properties to return a string
        String minLogReadActivityTextBTR = "test";
        when(properties.getMinLogReadActivityTextBTR()).thenReturn(minLogReadActivityTextBTR);

        // Act
        try {
            classUnderTest.getTreatmentWill(serviceActor, requestContext, null);
            fail();
        } catch (ServiceException e) {
            verify(accessControl).checkGetAccess(eq(requestContext));
            verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
            verify(treatmentWillValidator).validateTreatmentWillExists(eq(treatmentWillVO));
            verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER));

            assertEquals(FaultCode.GET_TREATMENT_WILL_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void hasTreatmentWill_true() throws DAOException {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();

        // Stub DAO method to return the TreatmentWillVO
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVO);

        // Act
        boolean hasTreatmentWill = classUnderTest.hasTreatmentWill(requestContext, null);

        // Assert
        verify(accessControl).checkHasAccess(eq(requestContext));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));

        assertTrue(hasTreatmentWill);
    }

    @Test
    public void hasTreatmentWill_false() throws DAOException {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample();

        // Stub DAO method to return null
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(null);

        // Act
        boolean hasTreatmentWill = classUnderTest.hasTreatmentWill(requestContext, null);

        // Assert
        verify(accessControl).checkHasAccess(eq(requestContext));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));

        assertFalse(hasTreatmentWill);
    }

    @Test
    public void createTreatmentWill_success() throws DAOException {
        // Arrange
        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        RequestContextVO requestContext = getRequestContextExample();
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return null (although the mock would do that anyway)
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class))).thenReturn(null);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class))).thenReturn(null);

        // Act
        classUnderTest.createTreatmentWill(treatmentWillVO, serviceActorExample, requestContext, null, false);

        // Assert
        verify(accessControl).checkCreateAccess(eq(requestContext));
        verify(livingWillDAO).getCurrentLivingWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(livingWillValidator).validateLivingWillNotExists(isNull());
        verify(treatmentWillValidator).validateTreatmentWillNotExists(isNull());
        verify(treatmentWillDAO).createTreatmentWill(eq(treatmentWillVO));
        assertEquals(treatmentWillVO.getPersonIdentifier(), treatmentWillVO.getPersonIdentifier());
        assertEquals(TreatmentWillServiceImpl.INITIAL_VERSION_DEFAULT, treatmentWillVO.getVersion().intValue());
        assertEquals(requestContext.getTime(), treatmentWillVO.getValidFrom());
        assertEquals(TreatmentWillServiceImpl.VALID_TO_DEFAULT, treatmentWillVO.getValidTo());
        assertEquals(requestContext.getTime(), treatmentWillVO.getModifiedDateTime());
    }

    @Test
    public void adminCreateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException {
        // Arrange
        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.Admin);
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return null (although the mock would do that anyway)
        when(livingWillDAO.getCurrentLivingWill(anyString(), any(Instant.class))).thenReturn(null);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class))).thenReturn(null);
        String minLogReadActivityTextBTR = "test";
        when(properties.getMinLogReadActivityTextBTR()).thenReturn(minLogReadActivityTextBTR);

        // Act
        classUnderTest.createTreatmentWill(treatmentWillVO, serviceActorExample, requestContext, null, false);

        // Assert
        verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER));
    }

    @Test
    public void updateTreatmentWill_success() throws DAOException {
        // Arrange
        long pid = 1L;
        int version = 1;

        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        RequestContextVO requestContext = getRequestContextExample();
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return an object with a PID and Version
        TreatmentWillVO treatmentWillVOFake = new TreatmentWillVO();
        treatmentWillVOFake.setTreatmentWillPID(pid);
        treatmentWillVOFake.setVersion(version);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVOFake);

        // Act
        classUnderTest.updateTreatmentWill(treatmentWillVO,serviceActorExample, requestContext, null, false);

        // Assert
        verify(accessControl).checkUpdateAccess(eq(requestContext));
        verify(treatmentWillDAO).getCurrentTreatmentWill(eq(PERSON_IDENTIFIER), eq(requestContext.getTime()));
        verify(treatmentWillValidator).validateTreatmentWillExists(eq(treatmentWillVOFake));
        verify(treatmentWillDAO).updateValidToTimestamp(eq(pid), eq(requestContext.getTime()));
        verify(treatmentWillDAO).createTreatmentWill(eq(treatmentWillVO));

        assertEquals(treatmentWillVO.getPersonIdentifier(), treatmentWillVO.getPersonIdentifier());
        assertEquals(TreatmentWillServiceImpl.INITIAL_VERSION_DEFAULT + 1, treatmentWillVO.getVersion().intValue());
        assertEquals(requestContext.getTime(), treatmentWillVO.getValidFrom());
        assertEquals(TreatmentWillServiceImpl.VALID_TO_DEFAULT, treatmentWillVO.getValidTo());
        assertEquals(requestContext.getTime(), treatmentWillVO.getModifiedDateTime());
    }

    @Test
    public void adminUpdateTreatmentWill_minLogCalled_success() throws DAOException, MinLogException {
        // Arrange
        long pid = 1L;
        int version = 1;

        TreatmentWillVO treatmentWillVO = getTreatmentWillExample();
        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.Admin);
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return an object with a PID and Version
        TreatmentWillVO treatmentWillVOFake = new TreatmentWillVO();
        treatmentWillVOFake.setTreatmentWillPID(pid);
        treatmentWillVOFake.setVersion(version);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVOFake);
        String minLogReadActivityTextBTR = "test";
        when(properties.getMinLogReadActivityTextBTR()).thenReturn(minLogReadActivityTextBTR);

        // Act
        classUnderTest.updateTreatmentWill(treatmentWillVO,serviceActorExample, requestContext, null, false);

        // Assert
        verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER));
    }

    @Test
    public void deleteTreatmentWill_success() throws DAOException {
        // Arrange
        long pid = 1L;
        int version = 1;

        RequestContextVO requestContext = getRequestContextExample();
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return an object with a PID and Version
        TreatmentWillVO treatmentWillVOFake = new TreatmentWillVO();
        treatmentWillVOFake.setTreatmentWillPID(pid);
        treatmentWillVOFake.setVersion(version);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVOFake);

        // Act
        classUnderTest.deleteTreatmentWill(serviceActorExample, requestContext, null);

        // Assert
        verify(accessControl).checkDeleteAccess(eq(requestContext));

        ArgumentCaptor<Long> updateValidToArgument1 = ArgumentCaptor.forClass(Long.class);
        ArgumentCaptor<Instant> updateValidToArgument2 = ArgumentCaptor.forClass(Instant.class);
        verify(treatmentWillDAO).updateValidToTimestamp(updateValidToArgument1.capture(), updateValidToArgument2.capture());

        assertEquals(pid, updateValidToArgument1.getValue().longValue());
        assertNotNull(updateValidToArgument2.getValue());
    }

    @Test
    public void adminDeleteTreatmentWill_minLogCalled_success() throws DAOException, MinLogException {
        // Arrange
        long pid = 1L;
        int version = 1;

        RequestContextVO requestContext = getRequestContextExample();
        requestContext.setUserType(UserType.Admin);
        ServiceActorVO serviceActorExample = getServiceActorExample();

        // Stub DAO method to return an object with a PID and Version
        TreatmentWillVO treatmentWillVOFake = new TreatmentWillVO();
        treatmentWillVOFake.setTreatmentWillPID(pid);
        treatmentWillVOFake.setVersion(version);
        when(treatmentWillDAO.getCurrentTreatmentWill(anyString(), any(Instant.class)))
                .thenReturn(treatmentWillVOFake);
        String minLogReadActivityTextBTR = "test";
        when(properties.getMinLogReadActivityTextBTR()).thenReturn(minLogReadActivityTextBTR);

        // Act
        classUnderTest.deleteTreatmentWill(serviceActorExample, requestContext, null);

        // Assert
        verify(minLogDataWrapper).logReadActivity(eq(requestContext), eq(minLogReadActivityTextBTR), anyString(), anyString(), anyString(), any(), eq(requestContext.getPersonIdentifier()), isNull(), eq(TreatmentWillServiceImpl.SERVICE_IDENTIFIER));
    }
}
