/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.dao;

import dk.sundhedsdatastyrelsen.livstestamente.service.AbstractSpringTest;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.LivingWillVO;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.time.Year;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;

import static org.junit.Assert.*;

@SpringBootTest
@Transactional("livingWillDataSourceTxManager")
public class LivingWillDAOImplTests extends AbstractSpringTest {

    private final Instant now = Instant.now();

    private final Integer version = 1;
    private final Instant validFrom = now;
    private final Instant validTo = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    private final String personIdentifier = "12345678";
    private final Instant modifiedDateTime = now;
    private final boolean noLifeProlongingIfDying = true;
    private final boolean noLifeProlongingIfSeverelyDegraded = true;

    @Autowired
    private LivingWillDAO livingWillDAO;

    @Test
    public void createAndGetCurrentLivingWill() throws Exception {
        // Arrange
        final LivingWillVO livingWillVO = createTestLivingWillVO();

        // Act
        final long pid = livingWillDAO.createLivingWill(livingWillVO);

        // Assert
        final LivingWillVO actualLivingWillVO = livingWillDAO.getCurrentLivingWill(personIdentifier, livingWillVO.getValidFrom());
        assertNotNull(actualLivingWillVO);
        assertTrue(pid >= 0);
        assertEquals(personIdentifier, actualLivingWillVO.getPersonIdentifier());
        assertEquals(version, actualLivingWillVO.getVersion());
        assertEquals(validFrom.truncatedTo(ChronoUnit.SECONDS), actualLivingWillVO.getValidFrom());
        assertEquals(validTo.truncatedTo(ChronoUnit.SECONDS), actualLivingWillVO.getValidTo());
        assertEquals(modifiedDateTime.truncatedTo(ChronoUnit.SECONDS), actualLivingWillVO.getModifiedDateTime());
        assertEquals(noLifeProlongingIfDying, actualLivingWillVO.isNoLifeProlongingIfDying());
        assertEquals(noLifeProlongingIfSeverelyDegraded, actualLivingWillVO.isNoLifeProlongingIfSeverelyDegraded());
    }

    @Test
    public void updateValidToTimeStamp() throws Exception {
        // Arrange
        final LivingWillVO livingWillVO = createTestLivingWillVO();
        final long pid = livingWillDAO.createLivingWill(livingWillVO);
        final Instant newValidTo = livingWillVO.getValidFrom().plusSeconds(1); // Add 1 second

        // Act
        livingWillDAO.updateValidToTimestamp(pid, newValidTo);

        // Assert
        final LivingWillVO actualLivingWillVO = livingWillDAO.getCurrentLivingWill(personIdentifier, livingWillVO.getValidFrom());
        assertNotNull(actualLivingWillVO);
        assertEquals(newValidTo.truncatedTo(ChronoUnit.SECONDS), actualLivingWillVO.getValidTo());
    }

    private LivingWillVO createTestLivingWillVO() {
        final LivingWillVO livingWillVO = new LivingWillVO();
        livingWillVO.setPersonIdentifier(personIdentifier);
        livingWillVO.setNoLifeProlongingIfDying(noLifeProlongingIfDying);
        livingWillVO.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegraded);
        livingWillVO.setVersion(version);
        livingWillVO.setModifiedDateTime(modifiedDateTime);
        livingWillVO.setValidFrom(validFrom);
        livingWillVO.setValidTo(validTo);
        return livingWillVO;
    }

}
