/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.dao;

import dk.sundhedsdatastyrelsen.livstestamente.service.AbstractSpringTest;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.AcceptanceNeeded;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.TreatmentWillVO;
import org.junit.Ignore;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.time.Year;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;

import static org.junit.Assert.*;

@SpringBootTest
@Transactional("treatmentWillDataSourceTxManager")
public class TreatmentWillDAOImplTests extends AbstractSpringTest {

    private final Instant now = Instant.now();

    private final Integer version = 1;
    private final Instant validFrom = now;
    private final Instant validTo = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    private final String personIdentifier = "12345678";
    private final Instant modifiedDateTime = now;
    private final boolean noLifeProlongingIfDying = true;
    private final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeeded = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;
    private final boolean noLifeProlongingIfSeverelyDegraded = true;
    private final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeeded = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;
    private final boolean noLifeProlongingIfSeverePain = true;
    private final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeeded = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;
    private final boolean noForcedTreatmentIfIncapable = true;
    private final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeeded = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;

    @Autowired
    private TreatmentWillDAO treatmentWillDAO;

    @Test
    public void createAndGetCurrentTreatmentWill() throws Exception {
        // Arrange
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();

        // Act
        final long pid = treatmentWillDAO.createTreatmentWill(treatmentWillVO);

        // Assert
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, treatmentWillVO.getValidFrom());
        assertNotNull(actualTreatmentWillVO);
        assertTrue(pid >= 0);
        assertEquals(personIdentifier, actualTreatmentWillVO.getPersonIdentifier());
        assertEquals(version, actualTreatmentWillVO.getVersion());
        assertEquals(validFrom.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidFrom());
        assertEquals(validTo.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidTo());
        assertEquals(modifiedDateTime.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getModifiedDateTime());

        assertEquals(noLifeProlongingIfDying, actualTreatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(noLifeProlongingIfDyingAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverelyDegraded, actualTreatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverePain, actualTreatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(noLifeProlongingIfSeverePainAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        assertEquals(noForcedTreatmentIfIncapable, actualTreatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(noForcedTreatmentIfIncapableAcceptanceNeeded, actualTreatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void createAndGetCurrentTreatmentWill_null() throws Exception {
        // Arrange
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();
        treatmentWillVO.setNoLifeProlongingIfDyingAcceptanceNeeded(null); // <--- is null
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded(null); // <--- is null
        treatmentWillVO.setNoLifeProlongingIfSeverePainAcceptanceNeeded(null); // <--- is null
        treatmentWillVO.setNoForcedTreatmentIfIncapableAcceptanceNeeded(null); // <--- is null

        // Act
        final long pid = treatmentWillDAO.createTreatmentWill(treatmentWillVO);

        // Assert
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, treatmentWillVO.getValidFrom());
        assertNotNull(actualTreatmentWillVO);
        assertTrue(pid >= 0);
        assertNull(actualTreatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded()); // <--- is null
        assertNull(actualTreatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded()); // <--- is null
        assertNull(actualTreatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded()); // <--- is null
        assertNull(actualTreatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded()); // <--- is null
    }

    @Ignore // disabled until a fix is implemented
    @Test
    public void createAndGetCurrentTreatmentWill_winterTime() throws Exception {
        // Arrange
        // Create a record a record with validFrom 10 minutes after the change to winter time (in Denmark)
        final Instant afterWinterTimeChange = Instant.parse("2018-10-28T00:10:00.000Z");

        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();
        treatmentWillVO.setValidFrom(afterWinterTimeChange);

        // Act
        treatmentWillDAO.createTreatmentWill(treatmentWillVO);

        // Assert
        // We should be able to find that record 55 minutes later
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(
                personIdentifier,
                afterWinterTimeChange.plus(55, ChronoUnit.MINUTES));
        assertNotNull(actualTreatmentWillVO);
        assertEquals(afterWinterTimeChange, actualTreatmentWillVO.getValidFrom());
    }

    @Test
    public void createAndGetCurrentTreatmentWill_summerTime() throws Exception {
        // Arrange
        // Create a record a record with validFrom 10 minutes after the change to summer time (in Denmark)
        final Instant afterSummerTimeChange = Instant.parse("2018-03-25T02:10:00.000Z");
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();
        treatmentWillVO.setValidFrom(afterSummerTimeChange);

        // Act
        treatmentWillDAO.createTreatmentWill(treatmentWillVO);

        // Assert
        // We should be able to find that record 55 minutes later
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(
                personIdentifier,
                afterSummerTimeChange.plus(55, ChronoUnit.MINUTES));
        assertNotNull(actualTreatmentWillVO);
        assertEquals(afterSummerTimeChange, actualTreatmentWillVO.getValidFrom());
    }

    @Test
    public void updateValidToTimeStamp() throws Exception {
        // Arrange
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();
        final long pid = treatmentWillDAO.createTreatmentWill(treatmentWillVO);
        final Instant newValidTo = treatmentWillVO.getValidFrom().plusSeconds(1); // Add 1 second

        // Act
        treatmentWillDAO.updateValidToTimestamp(pid, newValidTo);

        // Assert
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, treatmentWillVO.getValidFrom());
        assertNotNull(actualTreatmentWillVO);
        assertEquals(newValidTo.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidTo());
    }
    
    @Test
    public void createAndGetCurrentTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded() throws Exception {
    
        // Arrange
        final TreatmentWillVO treatmentWillVO = createTestTreatmentWillVO();

        // Act
        final long pid = treatmentWillDAO.createTreatmentWillSkipNoLifeProlongingIfDyingAcceptanceNeeded(treatmentWillVO);

        // Assert
        final TreatmentWillVO actualTreatmentWillVO = treatmentWillDAO.getCurrentTreatmentWill(personIdentifier, treatmentWillVO.getValidFrom());
        assertNotNull(actualTreatmentWillVO);
        assertTrue(pid >= 0);
        assertEquals(personIdentifier, actualTreatmentWillVO.getPersonIdentifier());
        assertEquals(version, actualTreatmentWillVO.getVersion());
        assertEquals(validFrom.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidFrom());
        assertEquals(validTo.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getValidTo());
        assertEquals(modifiedDateTime.truncatedTo(ChronoUnit.SECONDS), actualTreatmentWillVO.getModifiedDateTime());

        assertEquals(noLifeProlongingIfDying, actualTreatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(null, actualTreatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverelyDegraded, actualTreatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverePain, actualTreatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(noLifeProlongingIfSeverePainAcceptanceNeeded, actualTreatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        assertEquals(noForcedTreatmentIfIncapable, actualTreatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(noForcedTreatmentIfIncapableAcceptanceNeeded, actualTreatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    private TreatmentWillVO createTestTreatmentWillVO() {
        final TreatmentWillVO treatmentWillVO = new TreatmentWillVO();
        treatmentWillVO.setPersonIdentifier(personIdentifier);
        treatmentWillVO.setNoLifeProlongingIfDying(noLifeProlongingIfDying);
        treatmentWillVO.setNoLifeProlongingIfDyingAcceptanceNeeded(noLifeProlongingIfDyingAcceptanceNeeded);
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegraded);
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded(noLifeProlongingIfSeverelyDegradedAcceptanceNeeded);
        treatmentWillVO.setNoLifeProlongingIfSeverePain(noLifeProlongingIfSeverePain);
        treatmentWillVO.setNoLifeProlongingIfSeverePainAcceptanceNeeded(noLifeProlongingIfSeverePainAcceptanceNeeded);
        treatmentWillVO.setNoForcedTreatmentIfIncapable(noForcedTreatmentIfIncapable);
        treatmentWillVO.setNoForcedTreatmentIfIncapableAcceptanceNeeded(noForcedTreatmentIfIncapableAcceptanceNeeded);
        treatmentWillVO.setVersion(version);
        treatmentWillVO.setModifiedDateTime(modifiedDateTime);
        treatmentWillVO.setValidFrom(validFrom);
        treatmentWillVO.setValidTo(validTo);
        return treatmentWillVO;
    }

}
