/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.util;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.AssigningAuthorities;
import dk.sundhedsdatastyrelsen.livstestamente.OID;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.LivingWillService;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillService;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAMapper;
import org.hl7.btr.TreatmentWill;
import org.hl7.v3.POCDMT000040Entry;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.time.Instant;
import java.util.List;

@Component
public class TestHelper {

    public static final String ASSIGNING_AUTHORITY_NAME_PERSON = AssigningAuthorities.CPR;
    public static final String ORG_NAME = "OrgName";
    public static final String ORG_ID = OID.SOR;

    public static final String PERSON_IDENTIFIER = "0501792275";
    public static final String ADMIN_PERSON_IDENTIFIER = "18374828181";

    public static final boolean NO_LIFE_PROLONGING_IF_DYING = true;
    public static final AcceptanceNeeded NO_LIFE_PROLONGING_IF_DYING_ACCEPTANCE_NEEDED = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;

    public static final boolean NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED = true;
    public static final AcceptanceNeeded NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;

    public static final boolean NO_LIFE_PROLONGING_IF_SEVERE_PAIN = true;
    public static final AcceptanceNeeded NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;

    public static final boolean NO_FORCED_TREATMENT_IF_INCAPABLE = true;
    public static final AcceptanceNeeded NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED = AcceptanceNeeded.RELATIVE_ACCEPTANCE_REQUIRED;

    public ServiceActorVO getServiceActorExample() {
        ServiceActorVO serviceActorVO = new ServiceActorVO();
        serviceActorVO.setOrganizationId(ORG_ID);
        serviceActorVO.setOrganizationName(ORG_NAME);
        serviceActorVO.setPersonIdentifier(PERSON_IDENTIFIER);
        return serviceActorVO;
    }

    public RequestContextVO getRequestContextExample() {
        RequestContextVO requestContextVO = new RequestContextVO();
        requestContextVO.setTime(Instant.now());
        requestContextVO.setPersonIdentifier(PERSON_IDENTIFIER);
        requestContextVO.setUserType(UserType.DataOwner);
        return requestContextVO;
    }

    public LivingWillVO getLivingWillExample() {
        LivingWillVO livingWillVO = new LivingWillVO();
        livingWillVO.setPersonIdentifier(PERSON_IDENTIFIER);
        livingWillVO.setNoLifeProlongingIfDying(NO_LIFE_PROLONGING_IF_DYING);
        livingWillVO.setNoLifeProlongingIfSeverelyDegraded(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED);
        return livingWillVO;
    }

    public TreatmentWillVO getTreatmentWillExample() {
        TreatmentWillVO treatmentWillVO = new TreatmentWillVO();
        treatmentWillVO.setPersonIdentifier(PERSON_IDENTIFIER);
        treatmentWillVO.setNoLifeProlongingIfDying(NO_LIFE_PROLONGING_IF_DYING);
        treatmentWillVO.setNoLifeProlongingIfDyingAcceptanceNeeded(NO_LIFE_PROLONGING_IF_DYING_ACCEPTANCE_NEEDED);
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegraded(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED);
        treatmentWillVO.setNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED);
        treatmentWillVO.setNoLifeProlongingIfSeverePain(NO_LIFE_PROLONGING_IF_SEVERE_PAIN);
        treatmentWillVO.setNoLifeProlongingIfSeverePainAcceptanceNeeded(NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED);
        treatmentWillVO.setNoForcedTreatmentIfIncapable(NO_FORCED_TREATMENT_IF_INCAPABLE);
        treatmentWillVO.setNoForcedTreatmentIfIncapableAcceptanceNeeded(NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED);
        return treatmentWillVO;
    }

    public LivingWillVO getLivingWill(LivingWillService service, String personIdentifier) {
        RequestContextVO context = getRequestContextExample();
        context.setPersonIdentifier(personIdentifier);

        ServiceActorVO serviceActorExample = getServiceActorExample();
        serviceActorExample.setPersonIdentifier(personIdentifier);

        return service.getLivingWill(serviceActorExample, context, null);
    }

    public TreatmentWillVO getTreatmentWill(TreatmentWillService service, String personIdentifier) {
        RequestContextVO context = getRequestContextExample();
        context.setPersonIdentifier(personIdentifier);

        ServiceActorVO serviceActorExample = getServiceActorExample();
        serviceActorExample.setPersonIdentifier(personIdentifier);

        return service.getTreatmentWill(serviceActorExample, context, null);
    }

    @Component
    public class HL7Helper {

        @Autowired
        public TreatmentWillHL7CDAMapper treatmentWillHL7CDAMapper;

        private final dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.ObjectFactory factoryBTR = new dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.ObjectFactory();

        public List<POCDMT000040Entry> getAllEntries(GetTreatmentWillResponse response) {
            return response.getClinicalDocument().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries();
        }

        public TreatmentWill getTreatmentWillExample() {
            final TreatmentWill treatmentWill = TreatmentWillHL7CDAMapper.oBTR.createTreatmentWill();
            treatmentWill.setNoLifeProlongingIfDying(treatmentWillHL7CDAMapper.toTreatmentWillValueType(NO_LIFE_PROLONGING_IF_DYING, NO_LIFE_PROLONGING_IF_DYING_ACCEPTANCE_NEEDED));
            treatmentWill.setNoLifeProlongingIfSeverelyDegraded(treatmentWillHL7CDAMapper.toTreatmentWillValueType(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED));
            treatmentWill.setNoLifeProlongingIfSeverePain(treatmentWillHL7CDAMapper.toTreatmentWillValueType(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED));
            treatmentWill.setNoForcedTreatmentIfIncapable(treatmentWillHL7CDAMapper.toTreatmentWillValueType(NO_FORCED_TREATMENT_IF_INCAPABLE, NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED));
            return treatmentWill;
        }

        public CreateTreatmentWillRequest createCreateTreatmentWillRequest(String personIdentifier, TreatmentWill treatmentWill) {
            final CreateTreatmentWillRequest request = factoryBTR.createCreateTreatmentWillRequest();
            request.setId(treatmentWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
            request.setTreatmentWill(treatmentWill);
            return request;
        }

        public UpdateTreatmentWillRequest createUpdateTreatmentWillRequest(String personIdentifier, TreatmentWill treatmentWill) {
            final UpdateTreatmentWillRequest request = factoryBTR.createUpdateTreatmentWillRequest();
            request.setId(treatmentWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
            request.setTreatmentWill(treatmentWill);
            return request;
        }

        public DeleteTreatmentWillRequest createDeleteTreatmentWillRequest(String personIdentifier) {
            DeleteTreatmentWillRequest request = factoryBTR.createDeleteTreatmentWillRequest();
            request.setId(treatmentWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
            return request;
        }

        public GetTreatmentWillRequest createGetTreatmentWillRequest(String personIdentifier) {
            GetTreatmentWillRequest request = factoryBTR.createGetTreatmentWillRequest();
            request.setId(treatmentWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
            return request;
        }

        public HasTreatmentWillRequest createHasTreatmentWillRequest(String personIdentifier) {
            HasTreatmentWillRequest request = factoryBTR.createHasTreatmentWillRequest();
            request.setId(treatmentWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
            return request;
        }

    }

}
