/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sds.dgws_idws_proxy.security.ProxyData;
import dk.sds.dgws_idws_proxy.security.RequestType;
import dk.sundhedsdatastyrelsen.livstestamente.service.conf.TestDAOConfig;
import dk.sundhedsdatastyrelsen.livstestamente.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.web.server.LocalServerPort;
import org.springframework.oxm.jaxb.Jaxb2Marshaller;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.ws.client.core.WebServiceMessageCallback;
import org.springframework.ws.client.core.WebServiceOperations;
import org.springframework.ws.client.core.WebServiceTemplate;
import org.springframework.ws.soap.SoapMessage;
import org.springframework.ws.transport.context.TransportContext;
import org.springframework.ws.transport.context.TransportContextHolder;
import org.springframework.ws.transport.http.HttpUrlConnection;

import javax.servlet.ServletContext;

@RunWith(SpringRunner.class)
@SpringBootTest(webEnvironment = SpringBootTest.WebEnvironment.RANDOM_PORT)
public abstract class AbstractWebServiceTest {

    @MockBean
    public MinLogDataWrapper minLogDataWrapper;

    @Autowired
    TestHelper testHelper;

    @Autowired
    TestHelper.HL7Helper hl7Helper;

    @Autowired
    TestDAOConfig daoConfig;

    @Autowired
    ServletContext servletContext;

    @LocalServerPort
    int port;

    private WebServiceOperations webServiceTemplate;

    abstract WebServiceTemplate getWebServiceTemplate();

    AbstractWebServiceTest() {
        this.webServiceTemplate = getWebServiceTemplate();
    }

    WebServiceTemplate createWebServiceTemplate(String jaxbContextPath) {
        Jaxb2Marshaller jaxb2Marshaller = new Jaxb2Marshaller();
        jaxb2Marshaller.setContextPath(jaxbContextPath);

        WebServiceTemplate webServiceTemplate = new WebServiceTemplate();
        webServiceTemplate.setMarshaller(jaxb2Marshaller);
        webServiceTemplate.setUnmarshaller(jaxb2Marshaller);

        return webServiceTemplate;
    }

    Object sendRequest(String uri, String soapAction, Object request, ProxyData proxyData) {
        String proxyDataString = proxyDataToJson(proxyData);
        WebServiceMessageCallback requestCallback = message -> {
            TransportContext context = TransportContextHolder.getTransportContext();
            HttpUrlConnection connection = (HttpUrlConnection) context.getConnection();
            connection.addRequestHeader("X-Proxy-Data", proxyDataString);
            ((SoapMessage) message).setSoapAction(soapAction);
        };
        return webServiceTemplate.marshalSendAndReceive(uri, request, requestCallback);
    }

    private String proxyDataToJson(ProxyData headerInfo) {
    	String timeStampString = ",requestTimestamp= {seconds:" + valueOrNullString(Long.toString(headerInfo.getRequestTimestamp().getEpochSecond()) ) + 
    			",nanos:" + valueOrNullString(Integer.toString(headerInfo.getRequestTimestamp().getNano()) ) +"}";
        return "{requestId=" + quotedValueOrNullString(headerInfo.getRequestId()) +
        		timeStampString +
                ",requestType=" + quotedValueOrNullString(headerInfo.getRequestType().toString()) +
                ",messageId=" + quotedValueOrNullString(headerInfo.getMessageId()) +
                ",flowId=" + quotedValueOrNullString(headerInfo.getFlowId()) +
                ",idCardType=" + quotedValueOrNullString(headerInfo.getIdCardType() != null ? headerInfo.getIdCardType().toString() : null) +
                ",idCardID=" + quotedValueOrNullString(headerInfo.getIdCardID()) +
                ",idCardIssuer=" + quotedValueOrNullString(headerInfo.getIdCardIssuer()) +
                ",idCardVersion=" + quotedValueOrNullString(headerInfo.getIdCardVersion()) +
                ",authenticationLevel=" + quotedValueOrNullString(headerInfo.getAuthenticationLevel()) +
                ",cprnr=" + quotedValueOrNullString(headerInfo.getCprnr()) +
                ",givenName=" + quotedValueOrNullString(headerInfo.getGivenName()) +
                ",surName=" + quotedValueOrNullString(headerInfo.getSurName()) +
                ",emailAddress=" + quotedValueOrNullString(headerInfo.getEmailAddress()) +
                ",role=" + quotedValueOrNullString(headerInfo.getRole()) +
                ",authorizationCode=" + quotedValueOrNullString(headerInfo.getAuthorizationCode()) +
                ",occupation=" + quotedValueOrNullString(headerInfo.getOccupation()) +
                ",itSystemName=" + quotedValueOrNullString(headerInfo.getItSystemName()) +
                ",careProviderID=" + quotedValueOrNullString(headerInfo.getCareProviderID()) +
                ",careProviderName=" + quotedValueOrNullString(headerInfo.getCareProviderName()) +
                ",careProviderType=" + quotedValueOrNullString(headerInfo.getCareProviderType()) + "}";
    }

    private String quotedValueOrNullString(String value) {
        return value != null ? "\"" + value + "\"" : "null";
    }
    private String valueOrNullString(String value) {
        return value != null ? value : "null";
    }

    private ProxyData getProxyDataForHealthCareProfessional(String personIdentifier, String authenticationLevel) {
        ProxyData proxyData = new ProxyData(RequestType.DGWS);
        proxyData.setCprnr(personIdentifier);
        proxyData.setAuthenticationLevel(authenticationLevel);
        return proxyData;
    }

    ProxyData getProxyDataForDataOwner(String personIdentifier) {
        ProxyData proxyData = new ProxyData(RequestType.IDWS);
        proxyData.setCprnr(personIdentifier);
        return proxyData;
    }

    protected ProxyData getProxyDataForAdminUser(String personIdentifier, String authenticationLevel) {
        ProxyData proxyData = new ProxyData(RequestType.DGWS);
        proxyData.setCprnr(personIdentifier);
        proxyData.setAuthenticationLevel(authenticationLevel);
        return proxyData;
    }
}
