/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sundhedsdatastyrelsen.livstestamente.OID;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.LivingWillServiceImpl;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.LivingWillHL7CDAValidator;
import org.flywaydb.core.Flyway;
import org.hl7.ltr.LivingWill;
import org.hl7.v3.ANY;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.ws.client.core.WebServiceTemplate;

import java.util.List;

import static dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper.*;
import static org.junit.Assert.*;

@SuppressWarnings("SameParameterValue")
public class LivingWillWebServiceTests extends AbstractWebServiceTest {

    @Autowired
    private LivingWillServiceImpl businessService;

    @Autowired
    private LivingWillHL7CDAMapper livingWillHL7CDAMapper;

    @Autowired
    private LivingWillHL7CDAValidator validator;

    private static boolean initialized = false;
    private Flyway flyway;
    private String uri;
    private String adminUri;
    private static final ObjectFactory factory = new ObjectFactory();
    private static final String jaxbContextPath = "dk.sundhedsdatastyrelsen.livstestamente._2018._05._01";

    @Before
    public void before() {
        flyway = daoConfig.getFlywayLivingWill();

        uri = "http://localhost:" + port + servletContext.getContextPath() + "/ltr";
        adminUri = "http://localhost:" + port + servletContext.getContextPath() + "/ltrAdmin";

        if (!initialized) {
            initialized = true;
            flyway.clean();
            flyway.migrate();
        }
    }

    @After
    public void after() {
        // Clean up database
        flyway.clean();
        flyway.migrate();
    }

    @Override
    WebServiceTemplate getWebServiceTemplate() {
        return createWebServiceTemplate(jaxbContextPath);
    }

    @Test
    public void getLivingWill_testResponseStructure() {
        // Arrange
        LivingWillVO livingWillExample = testHelper.getLivingWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createLivingWill(livingWillExample, serviceActorExample, requestContextExample, null);

        // Act
        final GetLivingWillRequest request = createGetLivingWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#GetLivingWill";
        final GetLivingWillResponse response = (GetLivingWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = getAllEntries(response);

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        // Cast the value element to a LivingWill-type
        LivingWill livingWillElement = (LivingWill) values.get(0);

        // Test values of LivingWill-element
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, livingWillElement.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, livingWillElement.isNoLifeProlongingIfSeverelyDegraded());
    }

    @Test
    public void adminGetLivingWill_testResponseStructure() {
        // Arrange
        LivingWillVO livingWillExample = testHelper.getLivingWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createLivingWill(livingWillExample, serviceActorExample, requestContextExample, null);

        // Act
        final GetLivingWillRequest request = createGetLivingWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#GetLivingWill";
        final GetLivingWillResponse response = (GetLivingWillResponse) sendRequest(adminUri, soapAction, request,
                getProxyDataForAdminUser(ADMIN_PERSON_IDENTIFIER, "4"));

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = getAllEntries(response);

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        // Cast the value element to a LivingWill-type
        LivingWill livingWillElement = (LivingWill) values.get(0);

        // Test values of LivingWill-element
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, livingWillElement.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, livingWillElement.isNoLifeProlongingIfSeverelyDegraded());
    }

    @Test
    public void getLivingWill_validateResponseStructure() {
        // Arrange
        LivingWillVO livingWillExample = testHelper.getLivingWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createLivingWill(livingWillExample, serviceActorExample, requestContextExample, null);

        // Act
        final GetLivingWillRequest request = createGetLivingWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#GetLivingWill";
        final GetLivingWillResponse response = (GetLivingWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert
        assertNotNull(response);

        validator.validateWithoutThrow(response);

        final int errorCount = validator.getValidationErrors().size();
        assertEquals(errorCount + " validation errors were found: " + validator.getCombinedErrorText(), 0, errorCount);
    }

    @Test
    public void hasLivingWill_testResponseStructure() {
        // Arrange
        LivingWillVO livingWillExample = testHelper.getLivingWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createLivingWill(livingWillExample, serviceActorExample, requestContextExample, null);

        // Act
        final HasLivingWillRequest request = createHasLivingWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#HasLivingWill";
        final HasLivingWillResponse hasLivingWillResponse = (HasLivingWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert
        assertNotNull(hasLivingWillResponse);
        assertTrue(hasLivingWillResponse.isWillExists());
    }

    @Test
    public void createLivingWill_testRequestMapping() {
        // Arrange
        final LivingWill livingWillExample = getLivingWillExample();

        // Act
        final CreateLivingWillRequest request = createCreateLivingWillRequest(PERSON_IDENTIFIER, livingWillExample);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#CreateLivingWill";
        final CreateLivingWillResponse createLivingWillResponse = (CreateLivingWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert using business layer
        assertNotNull(createLivingWillResponse);
        final LivingWillVO livingWillVO = testHelper.getLivingWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(livingWillVO);
        assertNotNull(livingWillVO.getLivingWillPID());
        assertEquals(1, livingWillVO.getVersion().intValue());
        assertTrue(livingWillVO.getValidFrom().isBefore(livingWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, livingWillVO.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, livingWillVO.isNoLifeProlongingIfSeverelyDegraded());
    }

    @Test
    public void adminCreateLivingWill_testRequestMapping() {
        // Arrange
        final LivingWill livingWillExample = getLivingWillExample();

        // Act
        final CreateLivingWillRequest request = createCreateLivingWillRequest(PERSON_IDENTIFIER, livingWillExample);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#CreateLivingWill";
        final CreateLivingWillResponse createLivingWillResponse = (CreateLivingWillResponse) sendRequest(adminUri, soapAction, request,
                getProxyDataForAdminUser(ADMIN_PERSON_IDENTIFIER, "4"));

        // Assert using business layer
        assertNotNull(createLivingWillResponse);
        final LivingWillVO livingWillVO = testHelper.getLivingWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(livingWillVO);
        assertNotNull(livingWillVO.getLivingWillPID());
        assertEquals(1, livingWillVO.getVersion().intValue());
        assertTrue(livingWillVO.getValidFrom().isBefore(livingWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, livingWillVO.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, livingWillVO.isNoLifeProlongingIfSeverelyDegraded());
    }

    @Test
    public void updateLivingWill_testRequestMapping() {
        // Arrange
        final boolean noLifeProlongingIfDyingUpdated = false;
        final boolean noLifeProlongingIfSeverelyDegradedUpdated = true;

        LivingWill livingWill = getLivingWillExample();
        livingWill.setNoLifeProlongingIfDying(noLifeProlongingIfDyingUpdated);
        livingWill.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegradedUpdated);

        LivingWillVO livingWillExample = testHelper.getLivingWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createLivingWill(livingWillExample, serviceActorExample, requestContextExample, null);

        // Act
        final UpdateLivingWillRequest request = createUpdateLivingWillRequest(PERSON_IDENTIFIER, livingWill);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#UpdateLivingWill";
        final UpdateLivingWillResponse updateLivingWillResponse = (UpdateLivingWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert using business layer
        assertNotNull(updateLivingWillResponse);
        final LivingWillVO livingWillVO = testHelper.getLivingWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(livingWillVO);
        assertNotNull(livingWillVO.getLivingWillPID());
        assertEquals(2, livingWillVO.getVersion().intValue());
        assertTrue(livingWillVO.getValidFrom().isBefore(livingWillVO.getValidTo()));
        assertEquals(noLifeProlongingIfDyingUpdated, livingWillVO.isNoLifeProlongingIfDying());
        assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, livingWillVO.isNoLifeProlongingIfSeverelyDegraded());
    }

    @Test
    public void adminUpdateLivingWill_testRequestMapping() {
        // Arrange
        final boolean noLifeProlongingIfDyingUpdated = false;
        final boolean noLifeProlongingIfSeverelyDegradedUpdated = true;

        LivingWill livingWill = getLivingWillExample();
        livingWill.setNoLifeProlongingIfDying(noLifeProlongingIfDyingUpdated);
        livingWill.setNoLifeProlongingIfSeverelyDegraded(noLifeProlongingIfSeverelyDegradedUpdated);

        LivingWillVO livingWillExample = testHelper.getLivingWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createLivingWill(livingWillExample, serviceActorExample, requestContextExample, null);

        // Act
        final UpdateLivingWillRequest request = createUpdateLivingWillRequest(PERSON_IDENTIFIER, livingWill);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#UpdateLivingWill";
        final UpdateLivingWillResponse updateLivingWillResponse = (UpdateLivingWillResponse) sendRequest(adminUri, soapAction, request,
                getProxyDataForAdminUser(ADMIN_PERSON_IDENTIFIER, "4"));

        // Assert using business layer
        assertNotNull(updateLivingWillResponse);
        final LivingWillVO livingWillVO = testHelper.getLivingWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(livingWillVO);
        assertNotNull(livingWillVO.getLivingWillPID());
        assertEquals(2, livingWillVO.getVersion().intValue());
        assertTrue(livingWillVO.getValidFrom().isBefore(livingWillVO.getValidTo()));
        assertEquals(noLifeProlongingIfDyingUpdated, livingWillVO.isNoLifeProlongingIfDying());
        assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, livingWillVO.isNoLifeProlongingIfSeverelyDegraded());
    }

    @Test
    public void deleteLivingWill_testRequestMapping() {
        // Arrange
        LivingWillVO livingWillExample = testHelper.getLivingWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createLivingWill(livingWillExample, serviceActorExample, requestContextExample, null);

        // Act
        final DeleteLivingWillRequest request = createDeleteLivingWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#DeleteLivingWill";
        final DeleteLivingWillResponse deleteLivingWillResponse = (DeleteLivingWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert using business layer
        // No data should be found
        assertNotNull(deleteLivingWillResponse);
        try {
            testHelper.getLivingWill(businessService, PERSON_IDENTIFIER);
            fail();
        } catch (ServiceException e) {
            assertEquals(FaultCode.LIVING_WILL_NO_DATA_FOUND, e.getFaultCode());
        }
    }

    @Test
    public void adminDeleteLivingWill_testRequestMapping() {
        // Arrange
        LivingWillVO livingWillExample = testHelper.getLivingWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createLivingWill(livingWillExample, serviceActorExample, requestContextExample, null);

        // Act
        final DeleteLivingWillRequest request = createDeleteLivingWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/livstestamente/2018/05/01#DeleteLivingWill";
        final DeleteLivingWillResponse deleteLivingWillResponse = (DeleteLivingWillResponse) sendRequest(adminUri, soapAction, request,
                getProxyDataForAdminUser(ADMIN_PERSON_IDENTIFIER, "4"));

        // Assert using business layer
        // No data should be found
        assertNotNull(deleteLivingWillResponse);
        try {
            testHelper.getLivingWill(businessService, PERSON_IDENTIFIER);
            fail();
        } catch (ServiceException e) {
            assertEquals(FaultCode.LIVING_WILL_NO_DATA_FOUND, e.getFaultCode());
        }
    }

    private CreateLivingWillRequest createCreateLivingWillRequest(String personIdentifier, LivingWill livingWill) {
        final CreateLivingWillRequest request = factory.createCreateLivingWillRequest();
        request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
        request.setLivingWill(livingWill);
        return request;
    }

    private UpdateLivingWillRequest createUpdateLivingWillRequest(String personIdentifier, LivingWill livingWill) {
        final UpdateLivingWillRequest request = factory.createUpdateLivingWillRequest();
        request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
        request.setLivingWill(livingWill);
        return request;
    }

    private DeleteLivingWillRequest createDeleteLivingWillRequest(String personIdentifier) {
        DeleteLivingWillRequest request = factory.createDeleteLivingWillRequest();
        request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
        return request;
    }

    private GetLivingWillRequest createGetLivingWillRequest(String personIdentifier) {
        GetLivingWillRequest request = factory.createGetLivingWillRequest();
        request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
        return request;
    }

    private HasLivingWillRequest createHasLivingWillRequest(String personIdentifier) {
        HasLivingWillRequest request = factory.createHasLivingWillRequest();
        request.setId(livingWillHL7CDAMapper.createId(OID.CPR, personIdentifier, TestHelper.ASSIGNING_AUTHORITY_NAME_PERSON));
        return request;
    }

    private LivingWill getLivingWillExample() {
        final LivingWill livingWill = LivingWillHL7CDAMapper.oLTR.createLivingWill();
        livingWill.setNoLifeProlongingIfDying(NO_LIFE_PROLONGING_IF_DYING);
        livingWill.setNoLifeProlongingIfSeverelyDegraded(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED);
        return livingWill;
    }

    private List<POCDMT000040Entry> getAllEntries(GetLivingWillResponse response) {
        return response.getClinicalDocument().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries();
    }
}
