/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.service.webservice;

import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.*;
import dk.sundhedsdatastyrelsen.livstestamente.service.business.TreatmentWillServiceImpl;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.AcceptanceNeeded;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.TreatmentWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.ServiceException;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAMapper;
import dk.sundhedsdatastyrelsen.livstestamente.service.webservice.hl7cda.TreatmentWillHL7CDAValidator;
import org.flywaydb.core.Flyway;
import org.hl7.btr.TreatmentWill;
import org.hl7.v3.ANY;
import org.hl7.v3.POCDMT000040Entry;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.ws.client.core.WebServiceTemplate;

import java.util.List;

import static dk.sundhedsdatastyrelsen.livstestamente.service.util.TestHelper.*;
import static org.junit.Assert.*;

@SuppressWarnings("SameParameterValue")
public class TreatmentWillWebServiceTests extends AbstractWebServiceTest {

    @Autowired
    private TreatmentWillServiceImpl businessService;

    @Autowired
    private TreatmentWillHL7CDAMapper treatmentWillHL7CDAMapper;

    @Autowired
    private TreatmentWillHL7CDAValidator validator;

    private static boolean initialized = false;
    private Flyway flyway;
    private String uri;
    private String adminUri;
    private static final String jaxbContextPath = "dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01";

    @Before
    public void before() {
        flyway = daoConfig.getFlywayTreatmentWill();

        uri = "http://localhost:" + port + servletContext.getContextPath() + "/btr";
        adminUri = "http://localhost:" + port + servletContext.getContextPath() + "/btrAdmin";

        if (!initialized) {
            initialized = true;
            flyway.clean();
            flyway.migrate();
        }
    }

    @After
    public void after() {
        // Clean up database
        flyway.clean();
        flyway.migrate();
    }

    @Override
    WebServiceTemplate getWebServiceTemplate() {
        return createWebServiceTemplate(jaxbContextPath);
    }

    @Test
    public void getTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();
        businessService.createTreatmentWill(treatmentWillExample, serviceActorExample, requestContextExample, null, false);

        // Act
        final GetTreatmentWillRequest request = hl7Helper.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#GetTreatmentWill";
        final GetTreatmentWillResponse response = (GetTreatmentWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = hl7Helper.getAllEntries(response);

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        // Cast the value element to a TreatmentWill-type
        TreatmentWill treatmentWillElement = (TreatmentWill) values.get(0);

        // Test values of TreatmentWill-element
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillElement.getNoLifeProlongingIfDying().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillElement.getNoLifeProlongingIfSeverePain().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
    }

    @Test
    public void adminGetTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();
        businessService.createTreatmentWill(treatmentWillExample, serviceActorExample, requestContextExample, null, false);

        // Act
        final GetTreatmentWillRequest request = hl7Helper.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#GetTreatmentWill";
        final GetTreatmentWillResponse response = (GetTreatmentWillResponse) sendRequest(adminUri, soapAction, request,
                getProxyDataForAdminUser(ADMIN_PERSON_IDENTIFIER, "4"));

        // Assert
        assertNotNull(response);
        final List<POCDMT000040Entry> allEntries = hl7Helper.getAllEntries(response);

        // Should be 1 entry element
        assertEquals(1, allEntries.size());

        // Should be 1 value element
        List<ANY> values = allEntries.get(0).getObservation().getValues();
        assertEquals(1, values.size());

        // Cast the value element to a TreatmentWill-type
        TreatmentWill treatmentWillElement = (TreatmentWill) values.get(0);

        // Test values of TreatmentWill-element
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillElement.getNoLifeProlongingIfDying().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillElement.getNoLifeProlongingIfSeverelyDegraded().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillElement.getNoLifeProlongingIfSeverePain().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillElement.getNoForcedTreatmentIfIncapable().isValue());
        assertEquals("relativeAcceptanceRequired", treatmentWillElement.getNoLifeProlongingIfDying().getAcceptanceNeeded());
    }

    @Test
    public void getTreatmentWill_validateResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createTreatmentWill(treatmentWillExample, serviceActorExample, requestContextExample, null, false);

        // Act
        final GetTreatmentWillRequest request = hl7Helper.createGetTreatmentWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#GetTreatmentWill";
        final GetTreatmentWillResponse response = (GetTreatmentWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert
        assertNotNull(response);

        validator.validateWithoutThrow(response);

        final int errorCount = validator.getValidationErrors().size();
        assertEquals(errorCount + " validation errors were found: " + validator.getCombinedErrorText(), 0, errorCount);
    }

    @Test
    public void hasTreatmentWill_testResponseStructure() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createTreatmentWill(treatmentWillExample, serviceActorExample, requestContextExample, null, false);

        // Act
        final HasTreatmentWillRequest request = hl7Helper.createHasTreatmentWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#HasTreatmentWill";
        final HasTreatmentWillResponse hasTreatmentWillResponse = (HasTreatmentWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert
        assertNotNull(hasTreatmentWillResponse);
        assertTrue(hasTreatmentWillResponse.isWillExists());
    }

    @Test
    public void createTreatmentWill_testRequestMapping() {
        // Arrange
        final TreatmentWill treatmentWillExample = hl7Helper.getTreatmentWillExample();

        // Act
        final CreateTreatmentWillRequest request = hl7Helper.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#CreateTreatmentWill";
        final CreateTreatmentWillResponse createTreatmentWillResponse = (CreateTreatmentWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert using business layer
        assertNotNull(createTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_DYING_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void adminCreateTreatmentWill_testRequestMapping() {
        // Arrange
        final TreatmentWill treatmentWillExample = hl7Helper.getTreatmentWillExample();

        // Act
        final CreateTreatmentWillRequest request = hl7Helper.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#CreateTreatmentWill";
        final CreateTreatmentWillResponse createTreatmentWillResponse = (CreateTreatmentWillResponse) sendRequest(adminUri, soapAction, request,
                getProxyDataForAdminUser(ADMIN_PERSON_IDENTIFIER, "4"));

        // Assert using business layer
        assertNotNull(createTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));
        assertEquals(NO_LIFE_PROLONGING_IF_DYING, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(NO_LIFE_PROLONGING_IF_DYING_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERELY_DEGRADED_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(NO_LIFE_PROLONGING_IF_SEVERE_PAIN_ACCEPTANCE_NEEDED, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(NO_FORCED_TREATMENT_IF_INCAPABLE_ACCEPTANCE_NEEDED, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void createTreatmentWill_testRequestMapping_null() {
        // Arrange
        final TreatmentWill treatmentWillExample = hl7Helper.getTreatmentWillExample();
        treatmentWillExample.getNoLifeProlongingIfDying().setAcceptanceNeeded(null); // <--- is null
        treatmentWillExample.getNoLifeProlongingIfSeverelyDegraded().setAcceptanceNeeded(null); // <--- is null
        treatmentWillExample.getNoLifeProlongingIfSeverePain().setAcceptanceNeeded(null); // <--- is null
        treatmentWillExample.getNoForcedTreatmentIfIncapable().setAcceptanceNeeded(null); // <--- is null

        // Act
        final CreateTreatmentWillRequest request = hl7Helper.createCreateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWillExample);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#CreateTreatmentWill";
        final CreateTreatmentWillResponse createTreatmentWillResponse = (CreateTreatmentWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert using business layer
        assertNotNull(createTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(1, treatmentWillVO.getVersion().intValue());
        assertNull(treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded()); // <--- is null
        assertNull(treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded()); // <--- is null
        assertNull(treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded()); // <--- is null
        assertNull(treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded()); // <--- is null
    }

    @Test
    public void updateTreatmentWill_testRequestMapping() {
        // Note: this also tests null values

        // Arrange
        final boolean noLifeProlongingIfDyingUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverelyDegradedUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverePainUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeededUpdated = null;
        final boolean noForcedTreatmentIfIncapableUpdated = false;
        final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeededUpdated = null;

        TreatmentWill treatmentWill = hl7Helper.getTreatmentWillExample();
        treatmentWill.setNoLifeProlongingIfDying(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfDyingUpdated, noLifeProlongingIfDyingAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverelyDegraded(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverelyDegradedUpdated, noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverePain(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverePainUpdated, noLifeProlongingIfSeverePainAcceptanceNeededUpdated));
        treatmentWill.setNoForcedTreatmentIfIncapable(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noForcedTreatmentIfIncapableUpdated, noForcedTreatmentIfIncapableAcceptanceNeededUpdated));

        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createTreatmentWill(treatmentWillExample, serviceActorExample, requestContextExample, null, false);

        // Act
        final UpdateTreatmentWillRequest request = hl7Helper.createUpdateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWill);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#UpdateTreatmentWill";
        final UpdateTreatmentWillResponse updateTreatmentWillResponse = (UpdateTreatmentWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert using business layer
        assertNotNull(updateTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(2, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));

        assertEquals(noLifeProlongingIfDyingUpdated, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(noLifeProlongingIfDyingAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverePainUpdated, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(noLifeProlongingIfSeverePainAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        assertEquals(noForcedTreatmentIfIncapableUpdated, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(noForcedTreatmentIfIncapableAcceptanceNeededUpdated, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void adminUpdateTreatmentWill_testRequestMapping() {
        // Note: this also tests null values

        // Arrange
        final boolean noLifeProlongingIfDyingUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfDyingAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverelyDegradedUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated = null;
        final boolean noLifeProlongingIfSeverePainUpdated = false;
        final AcceptanceNeeded noLifeProlongingIfSeverePainAcceptanceNeededUpdated = null;
        final boolean noForcedTreatmentIfIncapableUpdated = false;
        final AcceptanceNeeded noForcedTreatmentIfIncapableAcceptanceNeededUpdated = null;

        TreatmentWill treatmentWill = hl7Helper.getTreatmentWillExample();
        treatmentWill.setNoLifeProlongingIfDying(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfDyingUpdated, noLifeProlongingIfDyingAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverelyDegraded(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverelyDegradedUpdated, noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated));
        treatmentWill.setNoLifeProlongingIfSeverePain(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noLifeProlongingIfSeverePainUpdated, noLifeProlongingIfSeverePainAcceptanceNeededUpdated));
        treatmentWill.setNoForcedTreatmentIfIncapable(treatmentWillHL7CDAMapper.toTreatmentWillValueType(noForcedTreatmentIfIncapableUpdated, noForcedTreatmentIfIncapableAcceptanceNeededUpdated));

        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createTreatmentWill(treatmentWillExample, serviceActorExample, requestContextExample, null, false);

        // Act
        final UpdateTreatmentWillRequest request = hl7Helper.createUpdateTreatmentWillRequest(PERSON_IDENTIFIER, treatmentWill);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#UpdateTreatmentWill";
        final UpdateTreatmentWillResponse updateTreatmentWillResponse = (UpdateTreatmentWillResponse) sendRequest(adminUri, soapAction, request,
                getProxyDataForAdminUser(ADMIN_PERSON_IDENTIFIER, "4"));

        // Assert using business layer
        assertNotNull(updateTreatmentWillResponse);
        final TreatmentWillVO treatmentWillVO = testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
        assertNotNull(treatmentWillVO);
        assertNotNull(treatmentWillVO.getTreatmentWillPID());
        assertEquals(2, treatmentWillVO.getVersion().intValue());
        assertTrue(treatmentWillVO.getValidFrom().isBefore(treatmentWillVO.getValidTo()));

        assertEquals(noLifeProlongingIfDyingUpdated, treatmentWillVO.isNoLifeProlongingIfDying());
        assertEquals(noLifeProlongingIfDyingAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfDyingAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverelyDegradedUpdated, treatmentWillVO.isNoLifeProlongingIfSeverelyDegraded());
        assertEquals(noLifeProlongingIfSeverelyDegradedAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverelyDegradedAcceptanceNeeded());

        assertEquals(noLifeProlongingIfSeverePainUpdated, treatmentWillVO.isNoLifeProlongingIfSeverePain());
        assertEquals(noLifeProlongingIfSeverePainAcceptanceNeededUpdated, treatmentWillVO.getNoLifeProlongingIfSeverePainAcceptanceNeeded());

        assertEquals(noForcedTreatmentIfIncapableUpdated, treatmentWillVO.isNoForcedTreatmentIfIncapable());
        assertEquals(noForcedTreatmentIfIncapableAcceptanceNeededUpdated, treatmentWillVO.getNoForcedTreatmentIfIncapableAcceptanceNeeded());
    }

    @Test
    public void deleteTreatmentWill_testRequestMapping() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createTreatmentWill(treatmentWillExample, serviceActorExample, requestContextExample, null, false);

        // Act
        final DeleteTreatmentWillRequest request = hl7Helper.createDeleteTreatmentWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#DeleteTreatmentWill";
        final DeleteTreatmentWillResponse deleteTreatmentWillResponse = (DeleteTreatmentWillResponse) sendRequest(uri, soapAction, request, getProxyDataForDataOwner(PERSON_IDENTIFIER));

        // Assert using business layer
        // No data should be found
        assertNotNull(deleteTreatmentWillResponse);
        try {
            testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
            fail();
        } catch (ServiceException e) {
            assertEquals(FaultCode.TREATMENT_WILL_NO_DATA_FOUND, e.getFaultCode());
        }
    }

    @Test
    public void adminDeleteTreatmentWill_testRequestMapping() {
        // Arrange
        TreatmentWillVO treatmentWillExample = testHelper.getTreatmentWillExample();
        RequestContextVO requestContextExample = testHelper.getRequestContextExample();
        ServiceActorVO serviceActorExample = testHelper.getServiceActorExample();

        businessService.createTreatmentWill(treatmentWillExample, serviceActorExample, requestContextExample, null, false);

        // Act
        final DeleteTreatmentWillRequest request = hl7Helper.createDeleteTreatmentWillRequest(PERSON_IDENTIFIER);
        final String soapAction = "http://sundhedsdatastyrelsen.dk/behandlingstestamente/2018/05/01#DeleteTreatmentWill";
        final DeleteTreatmentWillResponse deleteTreatmentWillResponse = (DeleteTreatmentWillResponse) sendRequest(adminUri, soapAction, request,
                getProxyDataForAdminUser(ADMIN_PERSON_IDENTIFIER, "4"));

        // Assert using business layer
        // No data should be found
        assertNotNull(deleteTreatmentWillResponse);
        try {
            testHelper.getTreatmentWill(businessService, PERSON_IDENTIFIER);
            fail();
        } catch (ServiceException e) {
            assertEquals(FaultCode.TREATMENT_WILL_NO_DATA_FOUND, e.getFaultCode());
        }
    }
}
