/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.migration.business;

import de.siegmar.fastcsv.reader.CsvParser;
import de.siegmar.fastcsv.reader.CsvReader;
import de.siegmar.fastcsv.reader.CsvRow;
import dk.sundhedsdatastyrelsen.livstestamente.migration.dao.LivingWillMigrationDAO;
import dk.sundhedsdatastyrelsen.livstestamente.service.domain.LivingWillVO;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import org.apache.commons.io.input.BOMInputStream;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.time.DateTimeException;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.Year;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalAccessor;
import java.util.*;

@Component
@Scope("prototype")
public class CSVReader {
    private static final Logger logger = LogManager.getLogger(CSVReader.class);
    private static final Instant VALID_TO_DEFAULT = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    public int BATCH_SIZE = 20000;
    private int error = 0;
    private final DataValidator dataValidator;
    private final LivingWillMigrationDAO livingWillMigrationDAO;

    private int numberOfRowsInserted = 0;

    @Autowired
    public CSVReader(DataValidator dataValidator, LivingWillMigrationDAO livingWillMigrationDAO) {
        this.livingWillMigrationDAO = livingWillMigrationDAO;
        this.dataValidator = dataValidator;
    }

    public Map<String, String> readingPersonFileIteratively(File file) throws IOException, InterruptedException {
        Map<String, String> personMap = new HashMap<>();

        final FileInputStream fileInputStream = new FileInputStream(file);
        final BOMInputStream bomInputStream = new BOMInputStream(fileInputStream, false); // Do not include BOM
        final InputStreamReader inputStreamReader = new InputStreamReader(bomInputStream, StandardCharsets.UTF_8);

        CsvReader csvReader = new CsvReader();
        csvReader.setContainsHeader(true);
        csvReader.setTextDelimiter('\'');
        csvReader.setSkipEmptyRows(true);
        CsvParser csvParser = csvReader.parse(inputStreamReader);
        CsvRow row;
        boolean headersValidated = false;
        try {
            while ((row = csvParser.nextRow()) != null) {
                if (Thread.currentThread().isInterrupted())
                    throw new InterruptedException();

                if (!headersValidated) {
                    dataValidator.validateHeadersExists(csvParser.getHeader(), Arrays.asList("PersonID", "CivilRegistrationNumber"));
                    headersValidated = true;
                }

                try {
                    dataValidator.validatePersonData(row.getField("PersonID"), row.getField("CivilRegistrationNumber"));
                    personMap.put(row.getField("PersonID"), row.getField("CivilRegistrationNumber"));
                } catch (IllegalStateException e) {
                    error++;
                    logger.error("Manglende header i csv filen");
                } catch (IllegalArgumentException e) {
                    error++;
                    logger.error("Valideringsfejl: " + e.getMessage());
                }
            }
        } finally {
            inputStreamReader.close();
        }


        return personMap;
    }

    public void iterativeFileReading(File file, Map<String, String> personMap) throws IOException, InterruptedException, DAOException {
        final FileInputStream fileInputStream = new FileInputStream(file);
        final BOMInputStream bomInputStream = new BOMInputStream(fileInputStream, false); // Do not include BOM
        final InputStreamReader inputStreamReader = new InputStreamReader(bomInputStream, StandardCharsets.UTF_8);

        CsvReader csvReader = new CsvReader();
        csvReader.setContainsHeader(true);
        csvReader.setSkipEmptyRows(true);
        CsvParser csvParser = csvReader.parse(inputStreamReader);
        CsvRow row;
        boolean headersValidated = false;
        List<LivingWillVO> livingWillVOList = new ArrayList<>();

        try {
            while ((row = csvParser.nextRow()) != null) {
                if (Thread.currentThread().isInterrupted())
                    throw new InterruptedException();

                if (!headersValidated) {
                    dataValidator.validateHeadersExists(csvParser.getHeader(), Arrays.asList("PersonID", "Answer1",
                            "Answer2", "ValidDate", "EditDate"));
                    headersValidated = true;
                }

                try {
                    dataValidator.validate(personMap, row);
                    livingWillVOList.add(handleLoadedRows(row, personMap));
                    if (livingWillVOList.size() == BATCH_SIZE) {
                        createLivingWill(livingWillVOList);
                        this.numberOfRowsInserted += livingWillVOList.size();
                        livingWillVOList.clear();
                    }
                } catch (IllegalArgumentException e) {
                    error++;
                    logger.error("Valideringsfejl: " + e.getMessage());
                }
            }
        } finally {
            inputStreamReader.close();
        }

        if (livingWillVOList.size() > 0) {
            createLivingWill(livingWillVOList);
            this.numberOfRowsInserted += livingWillVOList.size();
            livingWillVOList.clear();
        }
    }

    private void createLivingWill(List<LivingWillVO> livingWillVOs) throws DAOException {
        livingWillMigrationDAO.createLivingWills(livingWillVOs);
    }

    private LivingWillVO handleLoadedRows(CsvRow row, Map<String, String> personMap) {
        LivingWillVO livingWillVO = new LivingWillVO();
        livingWillVO.setLivingWillPID(Long.parseLong(row.getField("PersonID")));
        livingWillVO.setPersonIdentifier(personMap.get(row.getField("PersonID")));
        livingWillVO.setNoLifeProlongingIfDying(createBooleanFormat(row.getField("Answer1")));
        livingWillVO.setNoLifeProlongingIfSeverelyDegraded(createBooleanFormat(row.getField("Answer2")));
        livingWillVO.setValidFrom(createDateFormat(row.getField("ValidDate")));
        livingWillVO.setModifiedDateTime(getModifiedDate(row.getField("EditDate")));
        livingWillVO.setValidTo(VALID_TO_DEFAULT);
        livingWillVO.setVersion(1);
        return livingWillVO;
    }

    private Boolean createBooleanFormat(String value) {
        if (value.equals("J")) {
            return true;
        } else if (value.equals("N")) {
            return false;
        }
        return null;
    }

    private Instant createDateFormat(String validDate) {
    	
        try {
            DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS");
            TemporalAccessor temporalAccessor = dateTimeFormatter.parse(validDate);
            LocalDateTime localDateTime = LocalDateTime.from(temporalAccessor);
            ZonedDateTime zonedDateTime = ZonedDateTime.of(localDateTime, ZoneId.systemDefault());
            Instant result = Instant.from(zonedDateTime).truncatedTo(ChronoUnit.SECONDS);
            return result;

        } catch (DateTimeException e) {
            logger.error("Fejl ved parse dato: " + validDate + " til format: yyyy-MM-dd HH:mm:ss.SSS");
        }
        return null;

    }

    private Instant getModifiedDate(String editDate) {

        Instant modifiedDateTime = null;
        try {
            DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS");
            TemporalAccessor temporalAccessor = dateTimeFormatter.parse(editDate);
            LocalDateTime localDateTime = LocalDateTime.from(temporalAccessor);
            ZonedDateTime zonedDateTime = ZonedDateTime.of(localDateTime, ZoneId.systemDefault());
            modifiedDateTime = Instant.from(zonedDateTime).truncatedTo(ChronoUnit.SECONDS);

        } catch (DateTimeException e) {
            modifiedDateTime = Instant.now().truncatedTo(ChronoUnit.SECONDS);
        }
        return modifiedDateTime;
    }
    
    public int getErrorCount() {
        return error;
    }

    public int getNumberOfRowsInserted() {
        return numberOfRowsInserted;
    }

    public void resetErrorCount() {
        this.error = 0;
    }
}
