/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.migration.business;

import de.siegmar.fastcsv.reader.CsvRow;
import dk.sundhedsdatastyrelsen.livstestamente.migration.util.ValidationErrors;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.List;
import java.util.Map;

@Component
public class DataValidator {

    public void validateHeadersExists(List<String> headersInFile, List<String> listOfRequiredHeaders) {
        if (!headersInFile.containsAll(listOfRequiredHeaders)) {
            throwException(format(ValidationErrors.HEADERS_NOT_VALID, String.join(",", headersInFile), String.join(",", listOfRequiredHeaders)));
        }
    }

    public void validatePersonData(String personId, String personIdentifier) {
        if (personId.isEmpty() || ! personId.chars().allMatch(Character::isDigit)) {
            throwException(format(ValidationErrors.PERSON_ID_INVALID, personId));
        }
        if (personIdentifier.isEmpty() || personIdentifier.length() != 10 || !personIdentifier.chars().allMatch(Character::isDigit)) {
            throwException(format(ValidationErrors.PERSON_IDENTIFIER_INVALID, personId, personIdentifier));
        }
    }

    public void validate(Map<String, String> personMap, CsvRow row) {
        try {
            validateltrMigrationData(personMap, row);
        } catch (Exception e) {
            throwException(e.getMessage());
        }
    }

    public boolean validateltrMigrationData(Map<String,String> personMap, CsvRow row) {
        boolean ok = true;
        String personId = row.getField("PersonID");
        ok &= validatePersonId(personMap, personId);

        ok &= validateBooleanType(row.getField("Answer1"), "Answer1", personId);
        ok &= validateBooleanType(row.getField("Answer2"), "Answer2", personId);
        ok &= validateDateType(row.getField("ValidDate"), "ValidDate", personId);

        return ok;
    }

    private boolean validatePersonId(Map<String, String> personMap, String personId) {
        if (personMap.containsKey(personId)) {
            return true;
        }
        throwException(format(ValidationErrors.PERSON_ID_NON_EXISTING, personId));
        return false;
    }

    private boolean validateBooleanType(String dataStr, String elementName, String personId) {
        if (dataStr == null) {
            return true;
        }

        if (dataStr.equals("J") || dataStr.equals("N")) {
            return true;
        }
        throwException(format(ValidationErrors.BOOLEAN_DATA_INVALID, personId, elementName, dataStr));
        return false;
    }

    private boolean validateDateType(String validDate, String elementName, String personId) {
        String dateFormat = "yyyy-MM-dd HH:mm:ss.SSS";
        if (validDate == null) {
            throwException(format(ValidationErrors.DATE_REQUIRED, personId, elementName));
        }
        
        try {
            DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern(dateFormat);
            LocalDateTime.parse(validDate, dateTimeFormatter);
        } catch (DateTimeParseException e) {
            throwException(format(ValidationErrors.DATE_INVALID, personId, elementName, dateFormat, validDate));
            return false;
        }
        return true;
    }

    public String format(String error, String... expressions) {
        if (expressions != null) {
            for (int i=0; i<expressions.length; i++) {
                if (expressions[i] == null) {
                    expressions[i] = "";
                }
            }
        }
        return String.format(error, expressions);
    }

    private void throwException(String validationErrors) {
        throw new IllegalArgumentException(validationErrors);
    }
}
