/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.migration.business;

import dk.sundhedsdatastyrelsen.livstestamente.migration.util.Properties;
import dk.sundhedsdatastyrelsen.livstestamente.migration.util.ValidationErrors;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.livstestamente.service.util.FaultUtil;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Map;

@Component
public class MigrationTaskHandler {
    private static final Logger logger = LogManager.getLogger(MigrationTaskHandler.class);

    private final CSVReader csvReader;
    private final Properties properties;

    @Autowired
    public MigrationTaskHandler(CSVReader csvReader, Properties properties) {
        this.csvReader = csvReader;
        this.properties = properties;
    }

    @Transactional
    public void runMigration() throws IOException, DAOException, InterruptedException {
        csvReader.resetErrorCount();
        ClassLoader classLoader = getClass().getClassLoader();
        URL personFileName = classLoader.getResource(properties.personFileName);
        URL ltrFileName = classLoader.getResource(properties.ltrFileName);
        if (personFileName == null || ltrFileName == null) {
            logger.error(ValidationErrors.NO_CSV_FILES);
            FaultUtil.throwFault(FaultCode.GENERAL_ERROR, "No csv files to migrate");
        }

        File personFile = new File(personFileName.getFile());
        File ltrFile = new File(ltrFileName.getFile());

        Map<String, String> personMap = csvReader.readingPersonFileIteratively(personFile);
        Thread.sleep(1);
        csvReader.iterativeFileReading(ltrFile, personMap);
    }

    public Integer getErrorCount() {
        return csvReader.getErrorCount();
    }

    public int getNumberOfRowsInserted() {
        return csvReader.getNumberOfRowsInserted();
    }
}
