/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.livstestamente.migration;

import de.siegmar.fastcsv.writer.CsvWriter;
import dk.sundhedsdatastyrelsen.livstestamente.migration.business.CSVReader;
import dk.sundhedsdatastyrelsen.livstestamente.migration.business.MigrationTaskHandler;
import dk.sundhedsdatastyrelsen.livstestamente.migration.util.ValidationErrors;
import dk.sundhedsdatastyrelsen.livstestamente.service.exception.DAOException;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.transaction.annotation.Transactional;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAccessor;
import java.util.*;
import java.util.concurrent.ThreadLocalRandom;

import static org.hamcrest.CoreMatchers.containsString;
import static org.junit.Assert.*;

@RunWith(SpringRunner.class)
@SpringBootTest
@Transactional
public class CSVReaderTest {
    final String personFile = "PersonData.csv";
    final String ltrFile = "LivingWillData.csv";

    List<String> personIdList = new ArrayList<>();
    final private String[] booleanTypes = {"J", "N"};
    Random rnd = new Random();

    @Autowired
    CSVReader csvReader;

    @Autowired
    MigrationTaskHandler migrationTaskHandler;

    @Before
    public void setUp() {
        csvReader.BATCH_SIZE = 10;
    }

    @Test
    public void csvReader_personData_loadedExpectedRows() throws IOException, InterruptedException {
        File personFile = new File(this.personFile);
        final int SIZE = 10;
        createPersonData(personFile, SIZE);
        Map<String, String> personMap = csvReader.readingPersonFileIteratively(personFile);
        assertNotNull(personMap);
        assertEquals(10, personMap.size());
        personFile.deleteOnExit();
    }

    @Test
    public void csvReader_ltrData_loadedExpectedRows() throws IOException, InterruptedException, DAOException {
        File personFile = new File(this.personFile);
        final int numberOfRegistrationsToCreate = 10;

        createPersonData(personFile, numberOfRegistrationsToCreate);
        File ltrFile = new File(this.ltrFile);
        createLtrData(ltrFile);

        Map<String, String> personMap = csvReader.readingPersonFileIteratively(personFile);
        csvReader.iterativeFileReading(ltrFile, personMap);
        assertEquals(numberOfRegistrationsToCreate, csvReader.getNumberOfRowsInserted());
        personFile.deleteOnExit();
        ltrFile.deleteOnExit();
    }

    @Test
    public void csvReader_invalidPersonHeader_returnsTheRightError() {
        File personFile = new File(this.personFile);
        final int SIZE = 10;

        createInvalidPersonData(personFile, SIZE);
        try {
            csvReader.readingPersonFileIteratively(personFile);
        } catch (Exception e) {
            checkError(e, ValidationErrors.HEADERS_NOT_VALID);
        }
        personFile.deleteOnExit();
    }

    @Test
    public void csvReader_invalidDataHeader_returnsTheRightError() {
        File personFile = new File(this.personFile);
        final int SIZE = 10;

        createPersonData(personFile, SIZE);
        File ltrFile = new File(this.ltrFile);
        createInvalidLtrData(ltrFile);

        try {
            Map<String, String> personMap = csvReader.readingPersonFileIteratively(personFile);
            csvReader.iterativeFileReading(ltrFile, personMap);
        } catch (Exception e) {
            checkError(e, ValidationErrors.HEADERS_NOT_VALID);
        }
        personFile.deleteOnExit();
        ltrFile.deleteOnExit();
    }

    private void createPersonData(File file, int SIZE)  {
        Collection<String[]> data = new ArrayList<>();
        data.add(new String[] { "PersonID", "CivilRegistrationNumber"});

        createData(SIZE, data);
        createFile(file, data);
    }

    private void createInvalidPersonData(File file, int SIZE) {
        Collection<String[]> data = new ArrayList<>();
        data.add(new String[] { "Id", "CivilRegistrationNumber"});
        createData(SIZE, data);
        createFile(file, data);
    }

    private void createData(int SIZE, Collection<String[]> data) {
        for (int i = 1; i <= SIZE; i++) {
            String personId = String.valueOf(i);
            String personIdentifier = getRandomCPRNumber();
            data.add(new String[]{personId, personIdentifier});
            personIdList.add(personId);
        }
    }

    private void createLtrData(File file)  {
        Collection<String[]> data = new ArrayList<>();
        data.add(new String[]{"PersonID", "Answer1", "Answer2", "ValidDate", "EditDate"});

        for (String personId : personIdList) {
            data.add(new String[]{personId, getRandomString(booleanTypes), getRandomString(booleanTypes),
                    getRandomDateStartingFrom("01/01/1970"), getRandomDateStartingFrom("01/01/2000")});
        }
        createFile(file, data);
    }

    private void createInvalidLtrData(File file)  {
        Collection<String[]> data = new ArrayList<>();
        data.add(new String[]{"PersonID", "Answer1", "Answer2", "ValidDate"});

        for (String personId : personIdList) {
            data.add(new String[]{personId, getRandomString(booleanTypes), getRandomString(booleanTypes),
                    getRandomDateStartingFrom("01/01/1970"), getRandomDateStartingFrom("01/01/2000")});
        }
        createFile(file, data);
    }

    private void createFile(File file, Collection<String[]> data)  {
        CsvWriter csvWriter = new CsvWriter();
        csvWriter.setLineDelimiter("\n".toCharArray());
        try {
            csvWriter.write(file, StandardCharsets.UTF_8, data);
        } catch (IOException e) {
            e.printStackTrace();
        }
        Assert.assertTrue(file.exists());
    }

    private String getRandomCPRNumber() {
        final String s = String.format("%02d%02d%02d%04d",
                getRandomIntBetween(1, 31),
                getRandomIntBetween(1, 12),
                getRandomIntBetween(0, 99),
                getRandomIntBetween(1000, 9999));
        return s;
    }

    private String getRandomDateStartingFrom(String date) {
    	DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("dd/MM/yyyy");
    	DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS");
    	TemporalAccessor temporalAccessor = dateFormatter.parse(date);
    	LocalDate dateStart = LocalDate.from(temporalAccessor);
    	LocalDateTime dateTimeEnd = LocalDateTime.now();
    	
    	ZoneId zoneId = ZoneId.systemDefault();
    	long dateStartLong = dateStart.atStartOfDay(zoneId).toInstant().toEpochMilli();
    	long dateTimeEndLong = dateTimeEnd.atZone(zoneId).toInstant().toEpochMilli();
    	
        long random = ThreadLocalRandom.current().nextLong(dateStartLong, dateTimeEndLong);
    	LocalDateTime randomDate = LocalDateTime.ofInstant(Instant.ofEpochMilli(random), ZoneId.systemDefault());
    	String randomDateString = dateTimeFormatter.format(randomDate);
        
        return randomDateString;
    }

    private String getRandomString(String[] arrayBetween) {
        int i = rnd.nextInt(arrayBetween.length);
        return (arrayBetween[i]);
    }

    private int getRandomIntBetween(int min, int max) {
        return rnd.nextInt(max - min + 1) + min;
    }

    private void checkError(Exception ex, String message) {
        String error = ex.toString();
        assertMessageContainsError(error, message);
    }

    private void assertMessageContainsError(String message, String error) {
        if (error.contains("%")) {
            String errorWithoutWildcard = error.substring(0, error.indexOf("%"));
            assertThat(message, containsString(errorWithoutWildcard));
        } else {
            assertThat(message, containsString(error));
        }
    }
}
